"""Implementation of Metropolis and UniformlyRandom kernels.

A simple Metropolis-Hastings kernel and a Multicell Metropolis kernel that can hop
among different supercell shapes.
"""

__author__ = "Luis Barroso-Luque"

from math import log

import numpy as np

from smol.moca.kernel.base import (
    ALL_BIAS,
    ALL_MCUSHERS,
    MCKernel,
    MulticellKernel,
    ThermalKernelMixin,
)


class MetropolisAcceptMixin:
    """Metropolis Criterion Mixin class.

    Simple mixin class impolementing the Metropolis-Hastings criterion for reause in
    MCKernels.

    Must be used in conjunction with inheritance from an MCKernel or derived class.
    """

    def _accept_step(self, occupancy, step):
        """Accept/reject a given step and populate trace and aux states accordingly.

        Args:
            occupancy (ndarray):
                Current occupancy
            step (Sequence of tuple):
                Sequence of tuples of ints representing an MC step
        """
        log_factor = self.mcusher.compute_log_priori_factor(occupancy, step)
        exponent = -self.beta * self.trace.delta_trace.enthalpy + log_factor

        if self.bias is not None:
            exponent += self.trace.delta_trace.bias

        self.trace.accepted = np.array(
            True if exponent >= 0 else exponent > log(self._rng.random())
        )
        return self.trace.accepted


class Metropolis(MetropolisAcceptMixin, ThermalKernelMixin, MCKernel):
    """A Metropolis-Hastings kernel.

    The classic and nothing but the classic.
    """

    valid_mcushers = ALL_MCUSHERS
    valid_bias = ALL_BIAS

    def __init__(
        self,
        ensemble,
        step_type,
        temperature,
        *args,
        seed=None,
        bias_type=None,
        bias_kwargs=None,
        **kwargs,
    ):
        """Initialize Metropolis Kernel.

        Args:
            ensemble (Ensemble):
                an Ensemble instance to obtain the features and parameters
                used in computing log probabilities.
            step_type (str):
                string specifying the MCMC step type.
            temperature (float):
                temperature at which the MCMC sampling will be carried out.
            args:
                positional arguments to instantiate the MCUsher for the
                corresponding step size.
            kwargs:
                keyword arguments to instantiate the MCUsher for the
                corresponding step size.
        """
        # order of arguments here matters based on the MRO, temperature must be first
        super().__init__(
            temperature,
            ensemble,
            step_type,
            *args,
            seed=seed,
            bias_type=bias_type,
            bias_kwargs=bias_kwargs,
            **kwargs,
        )


class MulticellMetropolis(MetropolisAcceptMixin, ThermalKernelMixin, MulticellKernel):
    """A Multicell Metropolis Hastings kernel.

    This is useful for example when one wants to sample over different supercell shapes,
    i.e. for generating special quasi-random structures.

    The ensembles in all kernels must have supercells of the same size, only the shape
    can change. All ensembles must be created from the same Hamiltonian. And all kernels
    must be of the same kind.

    There is no requirement for the individual kernels to have the same step type, bias,
    or any other setup parameter, which allows very flexible sampling strategies.
    However, users are responsible for ensuring that the overall sampling strategy is
    correct! (i.e. satisfies balance or detailed balance).
    """

    valid_mcushers = None
    valid_bias = None

    def __init__(
        self,
        mckernels,
        temperature,
        kernel_probabilities=None,
        kernel_hop_periods=5,
        kernel_hop_probabilities=None,
        seed=None,
        **kwargs,
    ):
        """Initialize MCKernel.

        Args:
            mckernels (list of MCKernels):
                a list of MCKernels instances to obtain the features and parameters
                used in computing log probabilities.
            temperature (float):
                Temperature that kernel hops are attempted at.
            seed (int): optional
                non-negative integer to seed the PRNG
            kernel_probabilities (Sequence of floats): optional
                Probability for choosing each of the supplied kernels
            kernel_hop_periods (int or Sequence of ints): optional
                number of steps between kernel hop attempts.
            kernel_hop_probabilities (Sequence of floats): optional
                probabilities for choosing each of the specified hop periods.
        """
        if not all(isinstance(kernel, Metropolis) for kernel in mckernels):
            raise ValueError("All kernels must be of type Metropolis")

        # order of arguments here matters based on the MRO, temperature must be first
        super().__init__(
            temperature,
            mckernels,
            kernel_probabilities=kernel_probabilities,
            kernel_hop_periods=kernel_hop_periods,
            kernel_hop_probabilities=kernel_hop_probabilities,
            seed=seed,
            **kwargs,
        )

    @ThermalKernelMixin.temperature.setter
    def temperature(self, temperature):
        """Set temperature and update all kernels."""
        self.trace.temperature = np.array(temperature, dtype=np.float64)
        self.beta = 1.0 / (self.kB * temperature)
        for kernel in self.mckernels:
            kernel.temperature = temperature



"""Container extension classes that hold C arrays of structs to ndarray pointers."""

__author__ = "Luis Barroso-Luque"


cimport cython
cimport numpy as np
from cpython.mem cimport PyMem_Free, PyMem_Malloc, PyMem_Realloc

from smol.utils.cluster.struct cimport (
    FloatArray1D,
    FloatArray2D,
    IntArray1D,
    IntArray2D,
    OrbitC,
)

# Is it possible to "template" for all array containers based on tempita


cdef class OrbitContainer:
    def __init__(self, tuple orbit_data, *args):
        """Python initialization

        Args:
            orbit_data (tuple):
                The orbit data should be given as a list of tuples where is tuple has
                the following information for the corresponding orbit:
                (orbit id, orbit bit_id, flattened correlation tensors, tensor indices)
        """
        # keep a python reference to the orbit_list so that it is not garbage collected
        self._orbit_data = orbit_data

    def __cinit__(self, tuple orbit_data, *args):
        self.size = len(orbit_data)
        self.data = <OrbitC*> PyMem_Malloc(self.size * sizeof(OrbitC))
        if not self.data:
            raise MemoryError()

        # populate orbits array
        self.set_orbits(orbit_data)

    def __reduce__(self):
        """Return a tuple of the arguments needed to re-initialize the object."""
        return OrbitContainer, (self._orbit_data,)

    cpdef public void set_orbits(self, tuple orbit_data) except *:
        """Populated data using a list of orbit data."""
        cdef int i

        # check that dataypes are correct
        for data in orbit_data:
            if not isinstance(data[0], int):
                raise TypeError("id must be an integer.")
            if not isinstance(data[1], int):
                raise TypeError("bit_id must be an integer.")
            if not isinstance(data[2], np.ndarray):
                raise TypeError("correlation_tensors must be a numpy array.")
            if not isinstance(data[3], np.ndarray):
                raise TypeError("tensor_indices must be a numpy array.")
            if data[2].ndim != 2:
                raise ValueError("correlation_tensors must be 2D.")
            if data[3].ndim != 1:
                raise ValueError("tensor_indices must be 1D.")

        # if different size then reallocate
        if len(orbit_data) != self.size:
            mem = <OrbitC*> PyMem_Realloc(
                self.data, len(orbit_data) * sizeof(OrbitC)
            )
            if not mem:
                raise MemoryError()

            self.size = len(orbit_data)
            self.data = mem

        for i, data in enumerate(orbit_data):
            self.data[i] = OrbitContainer.create_struct(
                data[0], data[1], data[2], data[3]
            )

        self._orbit_data = orbit_data

    @staticmethod
    cdef OrbitC create_struct(
            int orbit_id,
            int bit_id,
            np.float64_t[:, ::1] correlation_tensors,
            np.int32_t[::1] tensor_indices,
    ):
        """Set the fields of a OrbitC struct from memoryviews."""
        cdef OrbitC orbit

        orbit.id = orbit_id
        orbit.bit_id = bit_id

        orbit.correlation_tensors.size_r = correlation_tensors.shape[0]
        orbit.correlation_tensors.size_c = correlation_tensors.shape[1]
        orbit.correlation_tensors.data = &correlation_tensors[0, 0]

        orbit.tensor_indices.size = tensor_indices.shape[0]
        orbit.tensor_indices.data = &tensor_indices[0]

        return orbit

    def __len__(self):
        return self.size

    def __dealloc__(self):
        PyMem_Free(self.data)


@cython.final
cdef class FloatArray2DContainer:
    def __init__(self, tuple arrays):
        # keep a python reference to the arrays so that it is not garbage collected
        self._arrays = arrays

    def __cinit__(self, tuple arrays):
        self.size = len(arrays)
        self.data = <FloatArray2D*> PyMem_Malloc(self.size * sizeof(FloatArray2D))
        if not self.data:
            raise MemoryError()

        # populate orbits array
        self.set_arrays(arrays)

    def __reduce__(self):
        """Return a tuple of the arguments needed to re-initialize the object."""
        return FloatArray2DContainer, (self._arrays,)

    cpdef public void set_arrays(self, tuple arrays) except *:
        """Populated data using a list of 2D arrays."""
        cdef int i

        for array in arrays:
            if array.ndim != 2:
                raise ValueError("All arrays must be 2D.")

        # if different size then reallocate
        if len(arrays) != self.size:
            mem = <FloatArray2D*> PyMem_Realloc(
                self.data, len(arrays) * sizeof(FloatArray2D)
            )
            if not mem:
                raise MemoryError()
            self.size = len(arrays)
            self.data = mem

        for i in range(self.size):
            self.data[i] = FloatArray2DContainer.create_struct(arrays[i])

        self._arrays = arrays


    @staticmethod
    cdef FloatArray2D create_struct(np.float64_t[:, ::1] array):
        """Set the fields of a _FloatArray2D struct from memoryview."""
        cdef FloatArray2D array_struct
        array_struct.size_r = array.shape[0]
        array_struct.size_c = array.shape[1]
        array_struct.data = &array[0, 0]
        return array_struct

    def __len__(self):
        return self.size

    def __dealloc__(self):
        PyMem_Free(self.data)


@cython.final
cdef class FloatArray1DContainer:
    def __init__(self, tuple arrays):
        # keep a python reference to the arrays so that it is not garbage collected
        self._arrays = arrays

    def __cinit__(self, tuple arrays):
        self.size = len(arrays)
        self.data = <FloatArray1D*> PyMem_Malloc(self.size * sizeof(FloatArray1D))
        if not self.data:
            raise MemoryError()

        # populate orbits array
        self.set_arrays(arrays)

    def __reduce__(self):
        """Return a tuple of the arguments needed to re-initialize the object."""
        return FloatArray1DContainer, (self._arrays,)

    cpdef public void set_arrays(self, tuple arrays) except *:
        """Populated data using a list of 1D arrays."""
        cdef int i

        for array in arrays:
            if array.ndim != 1:
                raise ValueError("All arrays must be 1D.")

        # if different size then reallocate
        if len(arrays) != self.size:
            mem = <FloatArray1D*> PyMem_Realloc(
                self.data, len(arrays) * sizeof(FloatArray1D)
            )
            if not mem:
                raise MemoryError()
            self.size = len(arrays)
            self.data = mem

        for i in range(self.size):
            self.data[i] = FloatArray1DContainer.create_struct(arrays[i])

        self._arrays = arrays

    @staticmethod
    cdef FloatArray1D create_struct(np.float64_t[::1] array):
        """Set the fields of a FloatArray1D struct from memoryview."""
        cdef FloatArray1D array_struct
        array_struct.size = array.size
        array_struct.data = &array[0]
        return array_struct

    def __len__(self):
        return self.size

    def __dealloc__(self):
        PyMem_Free(self.data)


@cython.final
cdef class IntArray1DContainer:
    def __init__(self, tuple arrays):
        # keep a python reference to the arrays so that it is not garbage collected
        self._arrays = arrays

    def __cinit__(self, tuple arrays):
        self.size = len(arrays)
        self.data = <IntArray1D*> PyMem_Malloc(self.size * sizeof(IntArray1D))
        if not self.data:
            raise MemoryError()

        # populate orbits array
        self.set_arrays(arrays)

    def __reduce__(self):
        """Return a tuple of the arguments needed to re-initialize the object."""
        return IntArray1DContainer, (self._arrays,)

    cpdef public void set_arrays(self, tuple arrays) except *:
        """Populated data using a list of 1D arrays."""
        cdef int i

        for array in arrays:
            if array.ndim != 1:
                raise ValueError("All arrays must be 1D.")

        # if different size then reallocate
        if len(arrays) != self.size:
            mem = <IntArray1D*> PyMem_Realloc(
                self.data, len(arrays) * sizeof(IntArray1D)
            )
            if not mem:
                raise MemoryError()
            self.size = len(arrays)
            self.data = mem

        for i in range(self.size):
            self.data[i] = IntArray1DContainer.create_struct(arrays[i])

        self._arrays = arrays

    @staticmethod
    cdef IntArray1D create_struct(np.int32_t[::1] array):
        """Set the fields of a _FloatArray1D struct from memoryview."""
        cdef IntArray1D array_struct
        array_struct.size = array.size
        array_struct.data = &array[0]
        return array_struct

    def __len__(self):
        return self.size

    def __dealloc__(self):
        PyMem_Free(self.data)


@cython.final
cdef class IntArray2DContainer:
    def __init__(self, tuple arrays):
        # keep a python reference to the arrays so that it is not garbage collected
        self._arrays = arrays

    def __cinit__(self, tuple arrays):
        self.size = len(arrays)
        self.data = <IntArray2D*> PyMem_Malloc(self.size * sizeof(IntArray2D))
        if not self.data:
            raise MemoryError()

        # populate orbits array
        self.set_arrays(arrays)

    def __reduce__(self):
        """Return a tuple of the arguments needed to re-initialize the object."""
        return IntArray2DContainer, (self._arrays,)

    cpdef public void set_arrays(self, tuple arrays)  except *:
        """Populated data using a list of 2D arrays."""
        cdef int i

        for array in arrays:
            if array.ndim != 2:
                raise ValueError("All arrays must be 2D.")

        # if different size then reallocate
        if len(arrays) != self.size:
            mem = <IntArray2D*> PyMem_Realloc(
                self.data, len(arrays) * sizeof(IntArray2D)
            )
            if not mem:
                raise MemoryError()
            self.size = len(arrays)
            self.data = mem

        for i in range(self.size):
            self.data[i] = IntArray2DContainer.create_struct(arrays[i])

        self._arrays = arrays

    @staticmethod
    cdef IntArray2D create_struct(np.int32_t[:, ::1] array):
        """Set the fields of a _IntArray2D struct from memoryview."""
        cdef IntArray2D array_struct
        array_struct.size_r = array.shape[0]
        array_struct.size_c = array.shape[1]
        array_struct.data = &array[0, 0]
        return array_struct

    def __len__(self):
        return self.size

    def __dealloc__(self):
        PyMem_Free(self.data)



"""Solver class for the ground state problem of cluster expansion."""

from typing import List, NamedTuple, Union
from warnings import warn

import cvxpy as cp
import numpy as np
from cvxpy.constraints.constraint import Constraint
from numpy.typing import ArrayLike

from smol.capp.generate.groundstate.upper_bound.constraints import (
    get_composition_space_constraints,
    get_fixed_composition_constraints,
    get_normalization_constraints,
)
from smol.capp.generate.groundstate.upper_bound.objectives import (
    get_terms_from_chemical_potentials,
    get_terms_from_decomposition_processor,
    get_terms_from_ewald_processor,
    get_terms_from_expansion_processor,
)
from smol.capp.generate.groundstate.upper_bound.terms import (
    get_expression_and_auxiliary_from_terms,
)
from smol.capp.generate.groundstate.upper_bound.variables import (
    get_occupancy_from_variables,
    get_variables_from_sublattices,
)
from smol.moca.ensemble import Ensemble
from smol.moca.occu_utils import get_dim_ids_table, occu_to_counts
from smol.moca.processor import (
    ClusterDecompositionProcessor,
    ClusterExpansionProcessor,
    CompositeProcessor,
    EwaldProcessor,
)

__author__ = "Fengyu Xie"


class ProblemCanonicals(NamedTuple):
    """CVXpy Canonical objects representing the underlying optimization problem.

    Attributes:
        problem (cp.Problem):
            The cvxpy problem to solve.
        variables (cp.Variable):
            Variables corresponding to each species on each site to be optimized.
        variable_indices (cp.Variable):
            List of variable index corresponding to each active site index and species
            indices in its site space. Inactive or restricted sites will not have any
            variable, but be marked as -1 or -2. See groundstate.upper_bound.variables.
        auxiliary_variables (cp.Variable):
            Slack variables used to linearize polynomial pseudo-boolean objective
            terms.
        indices_in_auxiliary_products (list of lists of int):
            A list containing the indices of variables whose product equals to the
            corresponding auxiliary slack variable.
        objective_function (cp.Expression):
            Objective function to be minimized. Usually energy per super-cell.
        constraints (list of cp.Constraint):
            List of constraints.
        num_auxiliary_constraints(int):
            Number of slack linearize constraints. Slack constraints are always
            the last constraints in the list.
    """

    problem: cp.Problem
    variables: cp.Variable
    # Index of variables on each active site.
    variable_indices: List[List[int]]
    auxiliary_variables: cp.Variable
    indices_in_auxiliary_products: List[List[int]]
    objective_function: cp.Expression
    constraints: Union[List[Constraint], None]
    num_auxiliary_constraints: int


class PeriodicGroundStateSolver:
    """Class to solve for the ground state occupancy/structure of a given ensemble.

    This implementation solves for the periodic ground-state configuration for a
    fixed super-cell size stored as the supercell matrix in the ensemble's processor.
    Meaning that the solution corresponds only to an upper bound of the global ground
    state of the corresponding infinitely sized system.

    For more details on the nature of the global problem, i.e. finding the upper and the
    lower bounds, please see:

    https://doi.org/10.1103/PhysRevB.94.134424

    .. warning::
        The number of terms and constraints in underlying mixed-integer problem grow
        very rapidly with the size of the supercell, the number of allowed species,
        and the number of terms in the cluster expansion. As a result, solving for
        ground-states of large supercells and complex systems can be very
        time-consuming or even impossible, even using proprietary solvers and many
        cores. Make sure to check the number of terms and constraints in the problem
        and benchmark on a smaller supercell.

    """

    def __init__(
        self,
        ensemble: Ensemble,
        charge_balanced: bool = True,
        initial_occupancy: ArrayLike = None,
        fixed_composition: ArrayLike = None,
        other_constraints: list = None,
        term_coefficients_cutoff: float = 0.0,
        warm_start: bool = False,
        solver: str = "SCIP",
        solver_options: dict = None,
    ):
        """Initialize PeriodicGroundStateSolver.

        Args:
            ensemble (Ensemble):
                An ensemble to initialize the problem with. If you want to modify
                sub-lattices, please do that in ensemble before initializing the
                groundstate. Sub-lattices can not be modified after giving the ensemble!
            charge_balanced (bool): optional
                Whether to enforce the charge-balance constraint. Default is True.
            initial_occupancy (ArrayLike): optional
                An initial occupancy used to set the occupancy of manually restricted
                sites that may have more than one allowed species. Also used to set up
                the initial composition when solving in a canonical ensemble.
                Must be provided if any site has been manually restricted, or solving
                canonical ensemble.
            fixed_composition (ArrayLike): optional
                An array of floats corresponding to the "counts" format of composition
                in CompositionSpace. See moca.composition. Unit is per super-cell.
                If the ensemble is canonical (no chemical potential has been set), will
                need to constrain the ground-state problem to a fixed composition.
                If initial_occupancy is provided while fixed_composition is None,
                composition will be fixed to that of initial_occupancy.
                Will be ignored in a semi-grand ensemble.
                If any of initial_occupancy or fixed_composition is provided, you are
                fully responsible to ensure that they satisfy charge balance and other
                constraints.
            other_constraints (list): optional
                Other constraints to set in composition space. See moca.composition
                for detailed description.
                Note that constraints are now to be satisfied with the number of sites
                in sub-lattices of the supercell instead of the primitive cell.
            term_coefficients_cutoff (float): optional
                Minimum cutoff to the coefficient of terms in the final polynomial Boolean
                objective function.
                If the absolute value of a term is less than this cutoff, it will not be
                included in the optimization. If no cutoff is given, will include every term
                in the optimization.
            solver (str): optional
               Specify the groundstate used to solve the problem.
               SCIP is default because it does not require
               to linearize the problem before solving a polynomial
               form pseudo-boolean optimization.
               See cvxpy and PySciOpt documentations on how to correctly
               configure SCIP for cvxpy.
            warm_start (bool): optional
               Whether to use the previous solution as an initialization
               of the current problem solve. Default is False.
            solver_options (dict): optional
               Options to be passed into the groundstate when calling Problem.solve().
               See cvxpy and the specific groundstate documentations for detail.
        """
        self.ensemble = ensemble
        self.charge_balanced = charge_balanced
        self._structure = self.ensemble.processor.structure
        self._sublattices = self.ensemble.sublattices

        self._other_constraints = other_constraints
        if initial_occupancy is not None:
            # Enforce int32.
            self.initial_occupancy = np.array(initial_occupancy, dtype=np.int32)
        else:
            self.initial_occupancy = None

        # Can always be set, but will only be used in a canonical ensemble.
        if fixed_composition is not None:
            self.fixed_composition = np.array(fixed_composition, dtype=int)
        elif self.initial_occupancy is not None:
            n_dims = sum(len(s.species) for s in self.sublattices)
            dim_ids_table = get_dim_ids_table(self.sublattices, active_only=False)
            self.fixed_composition = occu_to_counts(
                self.initial_occupancy, n_dims, dim_ids_table
            )
        else:
            self.fixed_composition = None

        self.cutoff = term_coefficients_cutoff
        self.warm_start = warm_start
        self.solver = solver
        self.solver_options = solver_options or {}

        self._ground_state_solution = None
        self._ground_state_energy = None
        self._ground_state_occupancy = None
        self._ground_state_structure = None

        self._initialize_problem()

    def _initialize_problem(self):
        """Generate variables, objective and constraints."""
        sublattices = self.sublattices
        variables, variable_indices = get_variables_from_sublattices(
            sublattices, self.structure, self.initial_occupancy
        )

        # Add constraints.
        constraints = get_normalization_constraints(variables, variable_indices)
        # Canonical ensemble.
        if self.ensemble.chemical_potentials is None:
            constraints.extend(
                get_fixed_composition_constraints(
                    sublattices,
                    variables,
                    variable_indices,
                    self.structure,
                    self.fixed_composition,
                )
            )
        # Semi-grand ensemble.
        else:
            constraints.extend(
                get_composition_space_constraints(
                    sublattices,
                    variables,
                    variable_indices,
                    self.structure,
                    charge_balanced=self.charge_balanced,
                    other_constraints=self._other_constraints,
                )
            )

        def _handle_single_processor(proc):
            # Give energy terms in the expression.
            if isinstance(proc, ClusterExpansionProcessor):
                return get_terms_from_expansion_processor(
                    variable_indices,
                    expansion_processor=proc,
                )
            if isinstance(proc, ClusterDecompositionProcessor):
                return get_terms_from_decomposition_processor(
                    variable_indices,
                    decomposition_processor=proc,
                )
            if isinstance(proc, EwaldProcessor):
                return get_terms_from_ewald_processor(
                    variable_indices,
                    ewald_processor=proc,
                )
            raise NotImplementedError(
                f"Ground state upper-bound objective function"
                f" not implemented for processor type:"
                f" {type(proc)}!"
            )

        # Objective energy function.
        processor = self.ensemble.processor
        if isinstance(processor, CompositeProcessor):
            terms = []
            for p in processor.processors:
                terms.extend(_handle_single_processor(p))
        else:
            terms = _handle_single_processor(processor)
        # E - mu N for semi-grand ensemble.
        if self.ensemble.chemical_potentials is not None:
            # Chemical potential term already includes the "-" before mu N.
            terms.extend(
                get_terms_from_chemical_potentials(
                    variable_indices,
                    chemical_table=self.ensemble._chemical_potentials["table"],
                )
            )

        (
            objective_func,
            aux_variables,
            indices_in_aux_products,
            aux_constraints,
        ) = get_expression_and_auxiliary_from_terms(terms, variables, self.cutoff)
        n_aux_constraints = len(aux_constraints)
        constraints.extend(aux_constraints)

        problem = cp.Problem(cp.Minimize(objective_func), constraints)

        self._canonicals = ProblemCanonicals(
            problem=problem,
            variables=variables,
            variable_indices=variable_indices,
            auxiliary_variables=aux_variables,
            indices_in_auxiliary_products=indices_in_aux_products,
            objective_function=objective_func,
            constraints=constraints,
            num_auxiliary_constraints=n_aux_constraints,
        )

    @property
    def structure(self):
        """Supercell structure of the ensemble without any split.

        Returns:
            Structure.
        """
        return self._structure

    @property
    def sublattices(self):
        """Sub-lattices to build the problem on.

        Potentially contains split sub-lattices or manually restricted sites.
        Returns:
            List of Sublattice.
        """
        return self._sublattices

    @property
    def canonicals(self):
        """The named tuple containing cvxpy problem details.

        Returns:
            ProblemCanonicals.
        """
        return self._canonicals

    @property
    def problem(self):
        """The cvxpy problem for solving upper-bound of ground-state.

        Returns:
            cvxpy.Problem.
        """
        return self.canonicals.problem

    @property
    def variables(self):
        """CVXPY variables for species on each active site.

        Returns:
            cvxpy.Variable.
        """
        return self.canonicals.variables

    @property
    def variable_indices(self):
        """Indices of variables on each site if site is active.

        Returns:
            list of lists of int: each sub-list contains variable
            indices on an active site.
        """
        return self.canonicals.variable_indices

    @property
    def objective_function(self):
        """The objective function to be minimized.

        Usually total energy per super-cell.

        Returns:
            cvxpy.Expression.
        """
        return self.canonicals.objective_function

    @property
    def constraints(self):
        """Constraints to be satisfied in minimization.

        Must include normalization per site. May include
        charge balance and other constraints to composition.
        In canonical ensemble, must include a constraint to
        fix composition.

        Returns:
            List[cvxpy.Constraint]
        """
        return self.canonicals.constraints

    @property
    def auxiliary_variables(self):
        """Auxiliary variables for linearizing the problem.

        Returns:
            cp.Variable.
        """
        return self.canonicals.auxiliary_variables

    @property
    def indices_in_auxiliary_products(self):
        """Indices of variables in cluster terms corresponding to auxiliary variables.

        Returns:
            list of lists of int
        """
        return self.canonicals.indices_in_auxiliary_products

    def solve(self):
        """Solve the MIP problem.

        Returns:
            np.ndarray, float:
                values of boolean variables in the ground-state solution, and
                the minimized ground-state energy.
        """
        if (
            self._ground_state_solution is not None
            and self._ground_state_energy is not None
        ):
            warn("Ground state already solved before. Overwriting previous result!")
        self.problem.solve(
            solver=self.solver, warm_start=self.warm_start, **self.solver_options
        )
        if (
            self.variables.value is not None
            and self.objective_function.value is not None
        ):
            self._ground_state_solution = self.variables.value.astype(int)
            self._ground_state_energy = self.objective_function.value
        else:
            warn(
                "Ground state could not be solved! Try using another solver, or"
                " fine tune solver options!"
            )
            self._ground_state_solution = None
            self._ground_state_energy = None
        # Return Nothing.

    def reset(self):
        """Clear previous solutions and reinitialize the problem."""
        self._ground_state_solution = None
        self._ground_state_energy = None
        self._ground_state_occupancy = None
        self._ground_state_structure = None

        self._initialize_problem()

    def _raise_unsolved(self):
        if self._ground_state_solution is None or self._ground_state_energy is None:
            raise RuntimeError(
                "The ground state of the current system is not"
                " solved before or was not successful. Call"
                " self.solve()!"
            )

    @property
    def ground_state_solution(self):
        """Boolean variable values at the ground state.

        Returns:
            np.ndarray: the solution array in 0 and 1.
        """
        self._raise_unsolved()
        return self._ground_state_solution

    @property
    def ground_state_energy(self):
        """Energy per super-cell in the ground-state.

        Returns:
            float.
        """
        self._raise_unsolved()
        return self._ground_state_energy

    @property
    def ground_state_occupancy(self):
        """Encoded occupancy string of the ground-state.

        Returns:
            np.ndarray
        """
        self._raise_unsolved()
        if self._ground_state_occupancy is None:
            self._ground_state_occupancy = get_occupancy_from_variables(
                self.sublattices,
                self.ground_state_solution,
                self.variable_indices,
            )
        return self._ground_state_occupancy

    @property
    def ground_state_structure(self):
        """The ground state structure.

        Returns:
            Structure.
        """
        self._raise_unsolved()
        if self._ground_state_structure is None:
            self._ground_state_structure = (
                self.ensemble.processor.structure_from_occupancy(
                    self.ground_state_occupancy
                )
            )
        return self._ground_state_structure



"""Utilities for building smol with OpenMP. Borrowed mostly from astropy and sklearn."""

import glob
import os
import subprocess
import sys
import tempfile
import textwrap
import traceback
import warnings

from setuptools.command.build_ext import customize_compiler, new_compiler

OMP_FLAGS = {
    "msvc": ["/openmp"],
    "mingw32": ["-fopenmp"],
    "other": ["-fopenmp"],
}


def compile_test_program(code, extra_preargs=None, extra_postargs=None):
    """Check that some C code can be compiled and run."""
    compiler = new_compiler()
    customize_compiler(compiler)

    start_dir = os.path.abspath(".")

    with tempfile.TemporaryDirectory() as tmp_dir:
        try:
            os.chdir(tmp_dir)

            # Write test program
            with open("test_program.c", "w") as f:
                f.write(code)

            os.mkdir("objects")
            # Compile, test program
            compiler.compile(
                ["test_program.c"], output_dir="objects", extra_postargs=extra_postargs
            )

            # Link test program
            objects = glob.glob(os.path.join("objects", "*" + compiler.obj_extension))
            compiler.link_executable(
                objects,
                "test_program",
                extra_preargs=extra_preargs,
                extra_postargs=extra_postargs,
            )

            if "PYTHON_CROSSENV" not in os.environ:
                # Run test program if not cross compiling
                # will raise a CalledProcessError if return code was non-zero
                output = subprocess.check_output("./test_program")
                output = output.decode(sys.stdout.encoding or "utf-8").splitlines()
            else:
                # Return an empty output if we are cross compiling
                # as we cannot run the test_program
                output = []
        except Exception:
            raise
        finally:
            os.chdir(start_dir)

    return output


def basic_check_build():
    """Check basic compilation and linking of C code."""
    code = textwrap.dedent(
        """\
        #include <stdio.h>
        int main(void) {
        return 0;
        }
        """
    )
    compile_test_program(code)


def get_openmp_flag(compiler):
    """Find the correct flag to enable OpenMP support."""
    if sys.platform.startswith("darwin") and "openmp" in os.getenv("CPPFLAGS", ""):
        # -fopenmp can't be passed as compile flag when using Apple-clang.
        # OpenMP support has to be enabled during preprocessing.
        #
        # For example, our macOS wheel build jobs use the following environment
        # variables to build with Apple-clang and the brew installed "libomp":
        #
        # export CPPFLAGS="$CPPFLAGS -Xpreprocessor -fopenmp"
        # export CFLAGS="$CFLAGS -I/usr/local/opt/libomp/include"
        # export CXXFLAGS="$CXXFLAGS -I/usr/local/opt/libomp/include"
        # export LDFLAGS="$LDFLAGS -Wl,-rpath,/usr/local/opt/libomp/lib
        #                          -L/usr/local/opt/libomp/lib -lomp"
        return []
    else:
        return OMP_FLAGS.get(compiler.compiler_type, OMP_FLAGS["other"])


def check_openmp_support(compiler):
    """Check whether OpenMP test code can be compiled and run."""
    code = textwrap.dedent(
        """\
        #include <omp.h>
        #include <stdio.h>
        int main(void) {
        #pragma omp parallel
        printf("nthreads=%d\\n", omp_get_num_threads());
        return 0;
        }
        """
    )
    extra_preargs = os.getenv("LDFLAGS", None)
    if extra_preargs is not None:
        extra_preargs = extra_preargs.strip().split(" ")
        # FIXME: temporary fix to link against system libraries on linux
        # "-Wl,--sysroot=/" should be removed
        extra_preargs = [
            flag
            for flag in extra_preargs
            if flag.startswith(("-L", "-Wl,-rpath", "-l"))
        ]

    extra_postargs = get_openmp_flag(compiler)

    try:
        output = compile_test_program(
            code, extra_preargs=extra_preargs, extra_postargs=extra_postargs
        )

        if output and "nthreads=" in output[0]:
            nthreads = int(output[0].strip().split("=")[1])
            openmp_supported = len(output) == nthreads
        elif "PYTHON_CROSSENV" in os.environ:
            # Since we can't run the test program when cross-compiling
            # assume that openmp is supported if the program can be
            # compiled.
            openmp_supported = True
        else:
            openmp_supported = False

    except Exception:
        # We could be more specific and only catch: CompileError, LinkError,
        # and subprocess.CalledProcessError.
        # setuptools introduced CompileError and LinkError, but that requires
        # version 61.1. Even the latest version of Ubuntu (22.04LTS) only
        # ships with 59.6. So for now we catch all exceptions and reraise a
        # generic exception with the original error message instead:
        openmp_supported = False
        traceback.print_exc()

    if not openmp_supported:
        message = textwrap.dedent(
            """

                            ***********
                            * WARNING *
                            ***********

            It seems that smol cannot be built with OpenMP.

            - See exception traceback for more details.

            - Make sure you have followed the installation instructions:

                https://cedergrouphub.github.io/smol/getting_started.html

            - If your compiler supports OpenMP but you still see this
              message, please submit a bug report at:

                https://github.com/CederGroupHub/smol/issues

            - The build will continue with OpenMP-based parallelism
              disabled. Note however that computing correlation and cluster interaction
              vectors will run in sequential mode instead of leveraging thread-based
              parallelism.

                                ***\n
            """
        )
        warnings.warn(message)

    return openmp_supported


def cythonize_extensions(extensions, **cythonize_kwargs):
    """Cythonize extensions."""
    import numpy
    from Cython.Build import cythonize

    # Fast fail before cythonization if compiler fails compiling basic test
    # code even without OpenMP
    basic_check_build()

    compiler_directives = {
        "language_level": 3,
        "boundscheck": False,  # set to True for debugging
        "nonecheck": False,
        "wraparound": False,
        "initializedcheck": False,
        "cdivision": True,
        **cythonize_kwargs,
    }

    return cythonize(
        extensions,
        include_path=[numpy.get_include()],
        compiler_directives=compiler_directives,
    )



"""Definitions of array and orbitC struct containers as cython extension types.

"""

__author__ = "Luis Barroso-Luque"

cimport numpy as np

import cython
from smol.utils.cluster.struct cimport (
    FloatArray1D,
    FloatArray2D,
    IntArray1D,
    IntArray2D,
    OrbitC,
)


cdef class OrbitContainer:
    cdef OrbitC* data
    cdef readonly int size
    cdef tuple _orbit_data

    cpdef public void set_orbits(self, tuple orbit_data) except *

    @staticmethod
    cdef OrbitC create_struct(
            int orbit_id,
            int bit_id,
            np.float64_t[:, ::1] correlation_tensors,
            np.int32_t[::1] tensor_indices,
    )


@cython.final
cdef class FloatArray2DContainer:
    cdef FloatArray2D* data
    cdef readonly int size
    cdef tuple _arrays

    cpdef public void set_arrays(self, tuple arrays) except *

    @staticmethod
    cdef FloatArray2D create_struct(np.float64_t[:, ::1] array)


@cython.final
cdef class FloatArray1DContainer:
    cdef FloatArray1D* data
    cdef readonly int size
    cdef tuple _arrays

    cpdef public void set_arrays(self, tuple arrays)  except *

    @staticmethod
    cdef FloatArray1D create_struct(np.float64_t[::1] array)


@cython.final
cdef class IntArray1DContainer:
    cdef IntArray1D* data
    cdef readonly int size
    cdef tuple _arrays

    cpdef public void set_arrays(self, tuple arrays)  except *

    @staticmethod
    cdef IntArray1D create_struct(np.int32_t[::1] array)


@cython.final
cdef class IntArray2DContainer:
    cdef IntArray2D* data
    cdef readonly int size
    cdef tuple _arrays

    cpdef public void set_arrays(self, tuple arrays)  except *

    @staticmethod
    cdef IntArray2D create_struct(np.int32_t[:, ::1] array)



"""Generate energy terms from processor to be converted into cvxpy."""

from copy import deepcopy
from itertools import product
from typing import List, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike

from smol.capp.generate.groundstate.upper_bound.indices import (
    map_ewald_indices_to_variable_indices,
)
from smol.moca.processor import (
    ClusterDecompositionProcessor,
    ClusterExpansionProcessor,
    EwaldProcessor,
)

__author__ = "Fengyu Xie"


def get_terms_from_expansion_processor(
    variable_indices: List[List[int]],
    expansion_processor: ClusterExpansionProcessor,
    group_output_by_function: bool = False,
) -> Union[
    List[Tuple[List[int], float, float]], List[List[Tuple[List[int], float, float]]]
]:
    """Get the cluster terms from cluster expansion processor.

    Args:
        variable_indices(list of lists of int):
            List of variable indices corresponding to each active site index and
            index of species in its site space. Inactive sites will be marked by
            either -1 or -2. See documentation in groundstate.upper_bound.variables.
        expansion_processor(ClusterExpansionProcessor):
            A cluster expansion processor to generate objective function with.
        group_output_by_function(bool): optional
            Whether to group the resulting terms by correlation functions, such
            that the output will be a list of lists, with each sub-list corresponding
            to a correlation function. Default is false.
            This is used specifically for checking the value of each correlation
            function in the objective function.
    Returns:
        list of tuples of (list of int, float, float):
            A list of tuples, each represents a cluster term in the energy
            representation, containing indices of variables to be taken product with,
            the correlation factor, and the corresponding coefficient.
            Energy is taken per super-cell.
            If group_output_by_function is enabled, the output will be a list of lists,
            with each sub-list corresponding to a correlation function.
    """
    coefs = expansion_processor.coefs
    # Store variable indices and the constant product in the cluster term.
    cluster_terms = [
        [] for _ in range(expansion_processor.cluster_subspace.num_corr_functions)
    ]
    cluster_terms[0] = [([], 1.0)]

    space = expansion_processor.cluster_subspace
    sc_matrix = expansion_processor.supercell_matrix
    mappings = space.supercell_orbit_mappings(sc_matrix)
    for orbit, mapping in zip(space.orbits, mappings):
        # Use un-flatten version now for easier access.
        n = deepcopy(orbit.bit_id)
        corr_tensors = orbit.correlation_tensors
        n_bit_combos = corr_tensors.shape[0]  # index of bit combos
        mapping = np.array(mapping, dtype=int)
        n_clusters = mapping.shape[0]  # cluster image index
        n_sites = mapping.shape[1]  # index within cluster
        for bid in range(n_bit_combos):
            for cid in range(n_clusters):
                # Get possible occupancy states for each site.
                site_states = []
                for sid_in_cluster in range(n_sites):
                    sid_in_supercell = mapping[cid, sid_in_cluster]
                    # Indices corresponding to un-split site space.
                    site_variable_indices = np.array(
                        variable_indices[sid_in_supercell], dtype=int
                    )
                    # -1 should never co-appear with non-negative variable index
                    # on a site.
                    # Some active variables exist.
                    if np.any(site_variable_indices >= 0):
                        site_states.append(
                            np.where(site_variable_indices >= 0)[0].tolist()
                        )
                    # Nothing active, then one must be -1. Can not have multiple -1.
                    else:
                        site_states.append(
                            np.where(site_variable_indices == -1)[0].tolist()
                        )

                # Only one of these cluster states can be active when a solution is valid.
                for cluster_state in product(*site_states):
                    # Tuple can be used to query a single element.
                    cluster_factor = corr_tensors[tuple((bid, *cluster_state))]
                    cluster_variable_indices = []
                    # Get index to query.
                    for sid_in_cluster in range(n_sites):
                        sid_in_supercell = mapping[cid, sid_in_cluster]
                        species_id_in_site = cluster_state[sid_in_cluster]
                        var_id = variable_indices[sid_in_supercell][species_id_in_site]
                        if var_id >= 0:
                            cluster_variable_indices.append(var_id)
                        else:
                            assert var_id == -1
                    # Divide by n_clusters to normalize.
                    cluster_terms[n].append(
                        [cluster_variable_indices, cluster_factor / n_clusters]
                    )
            n += 1

    # Put in system size and coefficients info.
    if not group_output_by_function:
        return [
            (inds, factor, coef * expansion_processor.size)
            for terms, coef in zip(cluster_terms, coefs)
            for inds, factor in terms
        ]
    else:
        return [
            [(inds, factor, coef * expansion_processor.size) for inds, factor in terms]
            for terms, coef in zip(cluster_terms, coefs)
        ]


def get_terms_from_decomposition_processor(
    variable_indices: List[List[int]],
    decomposition_processor: ClusterDecompositionProcessor,
    group_output_by_orbit: bool = False,
) -> Union[
    List[Tuple[List[int], float, float]], List[List[Tuple[List[int], float, float]]]
]:
    """Get the cluster terms from cluster decomposition processor.

    Args:
        variable_indices(list of lists of int):
            List of variable indices corresponding to each active site index and
            index of species in its site space. Inactive sites will be marked by
            either -1 or -2. See documentation in groundstate.upper_bound.variables.
        decomposition_processor(ClusterDecompositionProcessor):
            A cluster decomposition processor to generate objective function with.
        group_output_by_orbit(bool): optional
            Whether to group the resulting terms by orbits, such
            that the output will be a list of lists, with each sub-list corresponding
            to an orbit. Default is false.
            This is used specifically for checking the value of each interaction
            term in the objective function.
    Returns:
        list of tuples of (list of int, float, float):
            A list of tuples, each represents a cluster term in the energy
            representation, containing indices of variables to be taken product with,
            the interaction factor and the multiplicity coefficient.
            Energy is taken per super-cell.
            If group_output_by_orbit is enabled, the output will be a list of lists,
            with each sub-list corresponding to an orbit.
    """
    coefs = decomposition_processor.coefs  # Actually multiplicities.
    orbit_terms = [
        [] for _ in range(decomposition_processor.cluster_subspace.num_orbits)
    ]
    orbit_tensors = decomposition_processor._interaction_tensors
    # Use list in inner-layer as tuple does not support value assignment.
    orbit_terms[0] = [([], orbit_tensors[0])]

    n = 1
    space = decomposition_processor.cluster_subspace
    sc_matrix = decomposition_processor.supercell_matrix
    mappings = space.supercell_orbit_mappings(sc_matrix)
    for mapping in mappings:
        mapping = np.array(mapping, dtype=int)
        n_clusters = mapping.shape[0]  # cluster index.
        n_sites = mapping.shape[1]  # index within cluster.
        for cid in range(n_clusters):
            # Get possible occupancy states for each site.
            site_states = []
            for sid_in_cluster in range(n_sites):
                sid_in_supercell = mapping[cid, sid_in_cluster]
                # Indices corresponding to un-split site space.
                site_variable_indices = np.array(
                    variable_indices[sid_in_supercell], dtype=int
                )
                # -1 should never co-appear with non-negative variable index
                # on a site.
                # Some active variables exist.
                if np.any(site_variable_indices >= 0):
                    site_states.append(np.where(site_variable_indices >= 0)[0].tolist())
                # Nothing active, then one must be -1. Can not have multiple -1.
                else:
                    site_states.append(
                        np.where(site_variable_indices == -1)[0].tolist()
                    )

            # Only one of these cluster states can be active when a solution is valid.
            for cluster_state in product(*site_states):
                cluster_factor = orbit_tensors[n][tuple(cluster_state)]
                cluster_variable_indices = []
                # Get index to query.
                for sid_in_cluster in range(n_sites):
                    sid_in_supercell = mapping[cid, sid_in_cluster]
                    species_id_in_site = cluster_state[sid_in_cluster]

                    var_id = variable_indices[sid_in_supercell][species_id_in_site]
                    if var_id >= 0:
                        cluster_variable_indices.append(var_id)
                    else:
                        assert var_id == -1
                # Divide by n_clusters to normalize.
                orbit_terms[n].append(
                    [cluster_variable_indices, cluster_factor / n_clusters]
                )

        n += 1

    # Put in system size and coefficients info.
    if not group_output_by_orbit:
        return [
            (inds, factor, coef * decomposition_processor.size)
            for terms, coef in zip(orbit_terms, coefs)
            for inds, factor in terms
        ]
    else:
        return [
            [
                (inds, factor, coef * decomposition_processor.size)
                for inds, factor in terms
            ]
            for terms, coef in zip(orbit_terms, coefs)
        ]


def get_terms_from_ewald_processor(
    variable_indices: List[List[int]],
    ewald_processor: EwaldProcessor,
) -> List[Tuple[List[int], float, float]]:
    """Get the objective function from cluster expansion processor.

    Args:
        variable_indices(list of lists of int):
            List of variable indices corresponding to each active site index and
            index of species in its site space. Inactive sites will be marked by
            either -1 or -2. See documentation in groundstate.upper_bound.variables.
        ewald_processor(ClusterDecompositionProcessor):
            A cluster decomposition processor to generate objective function with.
            You are fully responsible for checking variables have the correct site
            orderings as will be in ewald_processor.
    Returns:
        list of tuples of (list of int, float, float):
            A list of tuples, each represents a cluster term in the energy
            representation, containing indices of variables to be taken product with,
            and the factors before the boolean product.
            Energy is taken per super-cell.
    """
    ewald_matrix = ewald_processor.ewald_matrix
    ewald_cluster_terms = []

    # Should be square matrix.
    n_ewald_rows = ewald_matrix.shape[0]

    ewald_to_var_ids = map_ewald_indices_to_variable_indices(
        ewald_processor.structure,
        variable_indices,
    )
    # Do not diagonalize because ewald matrix already contains 1/2.
    for i in range(n_ewald_rows):
        for j in range(n_ewald_rows):
            row_var_id = ewald_to_var_ids[i]
            col_var_id = ewald_to_var_ids[j]
            # Some species are never possible on some sites because of
            # manual restriction. Should skip any associated element.
            if row_var_id == -2 or col_var_id == -2:
                continue

            ewald_factor = ewald_matrix[i, j]
            ewald_variable_indices = []
            if row_var_id != -1:
                ewald_variable_indices.append(row_var_id)
            # Diagonal should also be a point term.
            if col_var_id != -1 and col_var_id != row_var_id:
                ewald_variable_indices.append(col_var_id)
            # Ewald processor coefficient is forced to be a 1D array.
            # Should convert it back.
            ewald_cluster_terms.append(
                (ewald_variable_indices, ewald_factor, float(ewald_processor.coefs))
            )

    # No need to multiply because system size is already included.
    return ewald_cluster_terms


def get_terms_from_chemical_potentials(
    variable_indices: List[List[int]],
    chemical_table: ArrayLike,
) -> List[Tuple[List[int], float, float]]:
    """Get the objective function from chemical potentials.

    Notice: returns the -mu N term. Negation already included.
    Args:
        variable_indices(list of lists of int):
            List of variable indices corresponding to each active site index and
            index of species in its site space. Inactive sites will be marked by
            either -1 or -2. See documentation in groundstate.upper_bound.variables.
        chemical_table(ArrayLike):
            Chemical potentials corresponding to each species (cols) on each site
            (rows). Simply use ensemble._chemical_potential["table"].
    Returns:
        list of tuples of (list of int, float):
            A list of tuples, each represents a cluster term in the energy
            representation, containing indices of variables to be taken product with,
            and the factor before the boolean product.
            Energy is taken per super-cell.
    """
    num_sites = len(variable_indices)

    point_terms = []
    for site_id in range(num_sites):
        # Assume table column indices match exactly with species codes.
        # Only an ensemble created from cluster expansion can satisfy that.
        for code, var_id in enumerate(variable_indices[site_id]):
            mu = chemical_table[site_id, code]
            # Multiply -1 to give E - mu * N.
            if var_id >= 0:
                point_terms.append(([var_id], -1 * mu, 1.0))
            elif var_id == -1:
                point_terms.append(([], -1 * mu, 1.0))

    # No need to multiply because system size already included.
    return point_terms



"""Tools for generation of fully random occupancies."""

__author__ = "Luis Barroso-Luque, Fengyu Xie"

import warnings

import numpy as np
from pymatgen.core import Composition, Element

from smol.cofe.space.domain import SiteSpace, Vacancy


# TODO alloy encoding/decoding and also allow it for initial occus in smapler, etc
def generate_random_ordered_occupancy(
    processor,
    composition=None,
    charge_neutral=False,
    tol=1e-6,
    encoded=True,
    rng=None,
    **kwargs,
):
    """Generate a random encoded occupancy according to a list of sublattices.

    Args:
        processor (Processor):
            A processor object that represents the supercell space.
        composition (Sequence of Composition): optional
            A sequence of pymatgen Compositions for each sublattice specifying the
            composition for the generated occupancy. Must be in the same order as the
            sublattices.
        charge_neutral (bool): optional
            If True, the generated occupancy will be charge neutral. Oxidation states
            must be present in sublattices, if a composition is given this option is
            ignored.
        tol (float): optional
            Tolerance for the composition check, only used if a composition is given.
        encoded (bool): optional
            If True then occupancy is given as integer array otherwise as a list of
            Species.
        rng (optional): {None, int, array_like[ints], SeedSequence, BitGenerator, Generator}
            A RNG, seed or otherwise to initialize defauly_rng

    Returns:
        ndarray: encoded occupancy
    """
    sublattices = processor.get_sublattices()

    if composition is None:
        if charge_neutral:
            occu = _gen_neutral_occu(sublattices, rng=rng, **kwargs)
        else:
            occu = _gen_unconstrained_ordered_occu(sublattices, rng=rng, **kwargs)
    else:
        occu = _gen_composition_ordered_occu(
            sublattices, composition, tol, rng=rng, **kwargs
        )

    if not encoded:
        occu = processor.decode_occupancy(occu)

    return occu


def _gen_unconstrained_ordered_occu(sublattices, rng=None):
    """Generate a random encoded occupancy according to a list of sublattices.

    Args:
        sublattices (Sequence of Sublattice):
            A sequence of sublattices
        rng (optional): {None, int, array_like[ints], SeedSequence, BitGenerator, Generator}
            A RNG, seed or otherwise to initialize defauly_rng

    Returns:
        ndarray: encoded occupancy
    """
    num_sites = sum(len(sl.sites) for sl in sublattices)
    rand_occu = np.zeros(num_sites, dtype=int)
    rng = np.random.default_rng(rng)
    for sublatt in sublattices:
        rand_occu[sublatt.sites] = rng.choice(
            sublatt.encoding, size=len(sublatt.sites), replace=True
        )
    # Enforce int32.
    return np.ascontiguousarray(rand_occu, dtype=np.int32)


def _gen_neutral_occu(sublattices, lam=10, num_attempts=10000, rng=None):
    """Generate a random encoded occupancy according to a list of sublattices.

    Args:
        sublattices (Sequence of Sublattice):
            A sequence of sublattices
        lam (float): optional
            TODO explanation...
        num_attempts (int): optional
            number of flip attempts to generate occupancy.
        rng (optional): {None, int, array_like[ints], SeedSequence, BitGenerator, Generator},
            A RNG, seed or otherwise to initialize defauly_rng

    Returns:
        ndarray: encoded occupancy
    """
    rng = np.random.default_rng(rng)

    def get_charge(sp):
        if isinstance(sp, (Element, Vacancy)):
            return 0
        else:
            return sp.oxi_state

    def charge(occu, sublattices):
        charge = 0
        for sl in sublattices:
            for site in sl.sites:
                sp_id = sl.encoding.tolist().index(occu[site])
                charge += get_charge(sl.species[sp_id])
        return charge

    def flip(occu, sublattices, lam=10):
        actives = [s for s in sublattices if s.is_active]
        sl = rng.choice(actives)
        site = rng.choice(sl.sites)
        code = rng.choice(list(set(sl.encoding) - {occu[site]}))
        occu_next = occu.copy()
        occu_next[site] = code
        C = charge(occu, sublattices)
        C_next = charge(occu_next, sublattices)
        accept = np.log(rng.random()) < -lam * (C_next**2 - C**2)
        if accept and C != 0:
            return occu_next.copy(), C_next
        else:
            return occu.copy(), C

    occu = _gen_unconstrained_ordered_occu(sublattices)
    for _ in range(num_attempts):
        occu, C = flip(occu, sublattices, lam=lam)
        if C == 0:
            # Enforce int32.
            return np.ascontiguousarray(occu, dtype=np.int32)

    raise TimeoutError(
        f"Can not generate a neutral occupancy in {num_attempts} attempts!"
    )


def _gen_composition_ordered_occu(sublattices, composition, tol, rng=None):
    """Generate a random occupancy satisfying a given composition.

    Args:
        sublattices (Sequence of Sublattice):
            A sequence of sublattices
        composition (Sequence of Composition): optional
            A sequence of pymatgen Compositions for each sublattice specifying the
            composition for the generated occupancy. Must be in the same order as the
            sublattices.
        tol (float):
            Tolerance for the composition check.
        rng (optional): {None, int, array_like[ints], SeedSequence, BitGenerator, Generator}
            A RNG, seed or otherwise to initialize defauly_rng

    Returns:
        ndarray: encoded occupancy
    """
    rng = np.random.default_rng(rng)
    compositions = _composition_compatiblity(sublattices, composition, tol, rng=rng)
    occu = np.zeros(sum(len(sl.sites) for sl in sublattices), dtype=int)

    for composition, sublattice in zip(compositions, sublattices):
        # create a dummy site space to account for vacancies
        composition = SiteSpace(composition)
        all_sites = list(sublattice.sites.copy())
        for sp, code in zip(sublattice.species, sublattice.encoding):
            num_sp = round(composition[sp] * len(sublattice.sites))
            sites = rng.choice(all_sites, size=num_sp, replace=False)
            occu[sites] = code
            all_sites = [i for i in all_sites if i not in sites]

    # Enforce int32.
    return np.ascontiguousarray(occu, dtype=np.int32)


def _composition_compatiblity(sublattices, composition, tol, rng=None):
    """Check if a composition is compatible with a list of sublattices.

    if only a single composition is given for more than one sublattice, a list of
    split compositions is returned.

    Args:
        sublattices (Sequence of Sublattice):
            A sequence of sublattices
        composition (Composition or Sequence of composition): optional
            A pymatgen Compositions that the generated occupancy should be. If a
            sequence of compositions is given the order must correspond to the order
            of sublattices.
        tol (float): optional
            Tolerance for the composition check.
        rng (optional): {None, int, array_like[ints], SeedSequence, BitGenerator, Generator}
            A RNG, seed or otherwise to initialize defauly_rng

    Returns:
        list of Composition: A list of compositions that are compatible
    """
    # check that all species in composition appear in sublattices
    if isinstance(composition, Composition):
        compositions = [composition]
    else:
        compositions = composition

    for i, (comp, sl) in enumerate(zip(compositions, sublattices)):
        if any(sp not in sl.site_space for sp in comp):
            raise ValueError(
                "species are present in composition that are not in sublattices."
            )
        if comp.num_atoms > 1:
            warnings.warn(
                "A given sublattice composition is not normalized. \n"
                "Will be turned to a fractional composition."
            )
            compositions[i] = comp.fractional_composition

    # check if the compositions are compatible with the size of the sublattices
    for composition, sublattice in zip(compositions, sublattices):
        total = 0
        for concentration in composition.values():
            num_sites = len(sublattice.sites) * concentration
            if abs(round(num_sites) - num_sites) > tol:
                raise ValueError("composition is not compatible with supercell size.")
            total += round(num_sites)

        if total > len(sublattice.sites) + tol:
            raise ValueError("composition is not compatible with supercell size.")

    return compositions



"""Get variables from processor."""

from typing import List, Tuple

import cvxpy as cp
import numpy as np
from numpy.typing import ArrayLike
from pymatgen.core import Structure

from smol.capp.generate.groundstate.upper_bound.indices import (
    get_sublattice_indices_by_site,
)
from smol.cofe.space.domain import get_allowed_species
from smol.moca.ensemble import Sublattice

__author__ = "Fengyu Xie"


def get_variables_from_sublattices(
    sublattices: List[Sublattice],
    structure: Structure,
    initial_occupancy: ArrayLike = None,
) -> Tuple[cp.Variable, List[List[int]]]:
    """Get cvxpy boolean variables for the upper-bound problem from a list sublattices.

    Inactive sites (sites will only 1 allowed species) will not be added into variables.
    Args:
        sublattices (list of Sublattice):
            Sub-lattices to build the upper-bound problem on.
            Note:1, Must contain all sub-lattices in the ensemble! 2, Must either be
            original sub-lattices computed from the structure, or sub-lattices
            split from those original sub-lattices with the initial_occupancy, such that
            the encoding of species can correctly refer to the species index in the
            structure's site_space!
        structure (Structure):
            The supercell structure stored in a processor's structure attribute that
            was used to create the given sub-lattices.
            The sub-lattices must match the processor structure, or they must be the
            result of splitting with the initial_occupancy. See smol.moca.sublattice for
            the explanation of splitting a sub-lattice.
        initial_occupancy (ArrayLike):
            An encoded occupancy string used to determine which species should occupy a
            manually restricted site in an active sub-lattice. If any site has been
            manually restricted, this argument will be mandatory.

    Returns:
        cp.Variable, list of lists of int:
          Flatten variables for each species on active sites; list of variable
          index corresponding to each site and species indices in
          its site space. Each sub-list will have the same shape as the corresponding
          original site space in structure.
          There are two cases when marking an inactive site:
          If a species is not allowed in the current sub-lattice setting (possibly due
          to a split sub-lattice does not contain this species), it will be marked as
          -2. If a species is enforced to occupy the site by manual site restriction,
          or due to the sub-lattice is inactive (i.e, sub-lattice only allows one
          species), then it will be marked by -1. In neither case a variable will be
          created on this site.
    """
    n_variables = 0
    num_sites = len(structure)
    # Original site spaces before any potential sub-lattice split and restriction.
    orig_site_spaces = get_allowed_species(structure)

    variable_indices = []
    site_sublattice_ids = get_sublattice_indices_by_site(sublattices)

    for site_id in range(num_sites):
        sublattice = sublattices[site_sublattice_ids[site_id]]
        current_site_space = sublattice.species
        orig_site_space = orig_site_spaces[site_id]
        # Only active sites are assigned with variables.
        site_variable_indices = []
        for species in orig_site_space:
            # Species in the site space after splitting.
            if species in current_site_space:
                # The current_site_space is active.
                if len(current_site_space) > 1:
                    # The current site is not manually restricted. Assign a variable.
                    if site_id in sublattice.active_sites:
                        site_variable_indices.append(n_variables)
                        n_variables += 1
                    # The current site is manually restricted. Need initial occupancy.
                    else:
                        if initial_occupancy is None:
                            raise ValueError(
                                f"Site {site_id} was manually restricted"
                                f" in sub-lattice: {sublattice}, but no initial"
                                f" occupancy was specified!"
                            )

                        # The species is exactly the enforced species as in
                        # initial_occupancy. Mark as -1 (always true).
                        # Encoding should match that of range(len(orig_site_space)).
                        if (
                            initial_occupancy[site_id]
                            == sublattice.encoding[current_site_space.index(species)]
                        ):
                            site_variable_indices.append(-1)
                        # The species is exactly the enforced species as in
                        # initial_occupancy. Mark as -2 (always false).
                        else:
                            site_variable_indices.append(-2)
                # The sub-lattice is inactive because it only has one species.
                # The site will always be occupied by this species. Mark as -1
                # (always true).
                else:
                    if len(current_site_space) == 0:
                        raise ValueError(
                            f"Got empty site space in sub-lattice" f" {sublattice}!"
                        )
                    site_variable_indices.append(-1)
            # Species no longer in site space after splitting. Mark as -2
            # (always false).
            else:
                site_variable_indices.append(-2)
        variable_indices.append(site_variable_indices)

    # Add a name to variable for better identification.
    return cp.Variable(n_variables, name="s", boolean=True), variable_indices


def get_occupancy_from_variables(
    sublattices: List[Sublattice],
    variable_values: ArrayLike,
    variable_indices: List[List[int]],
) -> np.ndarray:
    """Get encoded occupancy array from value of variables.

    Args:
        sublattices(list of Sublattice):
            Sub-lattices to build the upper-bound problem on.
        variable_values(ArrayLike):
            Value of cvxpy variables storing the ground-state result.
        variable_indices(list of lists of int):
            List of variable indices corresponding to each site index and
            the species in its site space.
    Returns:
        np.ndarray[np.int32]: Encoded occupancy string.
    """
    values = np.round(variable_values).astype(int)

    num_sites = len(variable_indices)
    # Enforce int32.
    occu = np.zeros(num_sites) - 1
    site_sublattice_ids = get_sublattice_indices_by_site(sublattices)

    # Not considering species encoding order yet.
    for site_id, indices in enumerate(variable_indices):
        sublattice = sublattices[site_sublattice_ids[site_id]]
        site_indices = np.array(indices, dtype=int)
        site_var_inds = site_indices[site_indices >= 0]
        # Active.
        if len(site_var_inds) > 0:
            species_ids_in_current_space = np.where(values[site_var_inds] == 1)[0]
            if len(species_ids_in_current_space) > 1:
                raise ValueError(f"More than one species occupied site {site_id}!")
            occu[site_id] = sublattice.encoding[species_ids_in_current_space[0]]
        # Inactive (manually restricted or site space is inactive.)
        # Simply set to the enforced code.
        else:
            occu[site_id] = np.where(site_indices == -1)[0][0]
    if np.any(occu < 0):
        raise ValueError(f"Variables does not match given indices: {variable_indices}!")

    # Enforce int32.
    return occu.astype(np.int32)


def get_variable_values_from_occupancy(
    sublattices: List[Sublattice],
    occupancy: ArrayLike,
    variable_indices: List[List[int]],
) -> np.ndarray:
    """Get value of variables from encoded occupancy array.

    Args:
        sublattices(list of Sublattice):
            Sub-lattices to build the upper-bound problem on.
        occupancy(ArrayLike): optional
            An encoded occupancy array. Does not check whether it satisfies
            constraints.
        variable_indices(list of lists of int):
            List of variable indices corresponding to each active site index and
            index of species in its site space. Inactive sites will be empty.

    Returns:
        np.ndarray: Value of boolean variables in 0 and 1.
    """
    # Variable indices are continuous.
    num_variables = max(max(sub) for sub in variable_indices) + 1
    values = np.zeros(num_variables, dtype=int)

    site_sublattice_ids = get_sublattice_indices_by_site(sublattices)

    for site_id, indices in enumerate(variable_indices):
        sublattice = sublattices[site_sublattice_ids[site_id]]
        site_indices = np.array(indices, dtype=int)
        site_var_inds = site_indices[site_indices >= 0]

        # Active site with variables assigned.
        if len(site_var_inds) > 0:
            active_var_id_in_current_space = np.where(
                occupancy[site_id] == sublattice.encoding
            )[0][0]
            active_var_id = site_var_inds[active_var_id_in_current_space]
            values[active_var_id] = 1

    # No check here, just return.
    return values



"""Defines a composition space with optional charge-neutrality or other constraints."""

__author__ = "Fengyu Xie"

import warnings
from itertools import chain

import numpy as np
from monty.json import MontyDecoder, MSONable
from pymatgen.core import Composition, Element

from smol.cofe.space.domain import Vacancy
from smol.moca.composition.constraints import CompositionConstraintsManager
from smol.moca.occu_utils import get_dim_ids_by_sublattice
from smol.utils.math import (
    NUM_TOL,
    get_ergodic_vectors,
    get_natural_centroid,
    get_natural_solutions,
    get_nonneg_float_vertices,
    get_optimal_basis,
    integerize_multiple,
    integerize_vector,
    solve_diophantines,
)


def get_oxi_state(sp):
    """Oxidation state from Specie/Element/Vacancy.

    Args:
       sp(Specie or Vacancy or Element):
          A species.
    Return:
       Charge of species: int.
    """
    if isinstance(sp, Element):
        return 0
    else:
        return int(sp.oxi_state)


def flip_vec_to_reaction(u, bits):
    """Convert flip direction into a reaction formula in string.

    This function is for easy interpretation of flip directions.
    Args:
        u(1D ArrayLike of int):
            The flip vector in number change of species.
        bits(list of lists of Specie or DummySpecie or Element or Vacancy):
            Species on all sub-lattices.
    Return:
        Reaction formulas: str.
    """
    u = np.array(u, dtype=int)
    dim_ids = get_dim_ids_by_sublattice(bits)

    from_strs = []
    to_strs = []

    for sl_id, (sl_species, sl_dims) in enumerate(zip(bits, dim_ids)):
        for specie, dim in zip(sl_species, sl_dims):
            if u[dim] < 0:
                from_strs.append(f"{-u[dim]} {str(specie)}({sl_id})")
            elif u[dim] > 0:
                to_strs.append(f"{u[dim]} {str(specie)}({sl_id})")

    from_str = " + ".join(from_strs)
    to_str = " + ".join(to_strs)
    return from_str + " -> " + to_str


class CompositionSpace(MSONable):
    """Composition space class.

    Generates a compositional space with given charge neutrality or additional
    composition constraints and a given a list of Species or DummySpecies and the number
    of sites in each sub-lattice in a given unit cell.

    A composition in the CompositionSpace can be expressed in 4 formats:
        1, Count of species on each sub-lattice, ordered and concatenated as
           each sub-lattice provided in self.bits and self.sublattice_sizes,
           into an 1D array. ("counts" format, also called unconstrained
           coordinates)
        2, Coordinates x on the constrained integer grid, computed with
           n = n0 + V @ x from its counts format representation n,
           where V are the grid basi vectors and n0 is a base solution in
           a certain super-cell size. ("coordinates" format, also called
           constrained coordinates).
        3, Composition of each sub-lattice in pymatgen.Composition,
           but each sub-lattice composition is required to be divided
           by number of the sites in the sub-lattice, so that the sum of species
           content can not exceed 1. Vacancies contents are not explicitly
           included, but taken a 1 - sum(other_species_contents).
           ("compositions" format).
        4, Count each species in the whole structure, without distinguishing
           sub-lattice. The species are sorted, and the order of species is
           given in self.species. ("species-counts" format).

    Attributes:
        num_dims (int):
            The dimension of unconstrained composition space, which equals to
            the sum of site space size in all sub-lattices. Namely, it is the
            dimension of "counts" format.
        species (List of Species or DummySpecies or Element or Vacancy):
            All species in the given system (sorted).
        dim_ids (list of lists of int):
            The corresponding index in the "counts" vector
            of each species on each sub-lattice.
        species_ids (list of lists of int):
            The index of sorted species in each sub-lattice as shown in self.species.
    """

    other_constraints = CompositionConstraintsManager()

    def __init__(
        self,
        site_spaces,
        sublattice_sizes=None,
        charge_neutral=True,
        other_constraints=None,
        optimize_basis=False,
        table_ergodic=False,
    ):
        """Initialize CompositionSpace.

        Args:
            site_spaces (list of lists of Specie or Vacancy or Element):
                Site spaces specifying the allowed species on each sub-lattice.
            sublattice_sizes (1D ArrayLike of int): optional
                Number of sites in each sub-lattice per primitive cell.
                If not given, assume one site for each sub-lattice.
                Better provide them as co-prime integers.
            charge_neutral (bool): optional
                Whether to add charge balance constraint. Default is true.
            other_constraints
            (list of tuples of (1D arrayLike[float], float, str) or str): optional
                Other composition constraints to be applied to restrict the
                enumerated compositions.
                Allows two formats for each constraint in the list:
                    1, A string that encodes the constraint equation.
                    For example: "2 Ag+(0) + Cl-(1) +3 H+(2) <= 3 Mn2+ +4".
                       A string representation of constraint must satisfy the following
                       rules,
                       a, Contains a relation symbol ("==", "<=", ">=" or "=") are
                       allowed.
                       The relation symbol must have exactly one space before and one
                       space after to separate the left and the right sides.
                       b, Species strings must be readable by get_species in smol.cofe
                       .space.domain. No space is allowed within a species string.
                       For the format of a legal species string, refer to
                       pymatgen.core.species and smol.cofe.
                       c, You can add a number in brackets following a species string
                       to specify constraining the amount of species in a particular
                       sub-lattice. If not given, will apply the constraint to this
                       species on all sub-lattices.
                       This sub-lattice index label must not be separated from
                       the species string with space or any other character.
                       d, Species strings along with any sub-lattice index label must
                       be separated from other parts (such as operators and numbers)
                       with at least one space.
                       e, The intercept terms (a number with no species that follows)
                       must always be written at the end on both side of the equation.
                    2, The equation expression, which is a tuple containing a list of
                    floats of length self.n_dims to give the left-hand side coefficients
                    of each component in the composition "counts" format, a float to
                    give the right-hand side, and a string to specify the comparative
                    relationship between the left- and right-hand sides. Constrained in
                    the form of a_left @ n = (or <= or >=) b_right.
                    The components in the left-hand side are in the same order as in
                    itertools.chain(*self.bits).
                Note that all numerical values in the constraints must be set as they are
                to be satisfied per primitive cell given the sublattice_sizes!
                For example, if each primitive cell contains 1 site in 1 sub-lattice
                specified as sublattice_sizes=[1], with the requirement that species
                A, B and C sum up to occupy less than 0.6 sites per sub-lattice, then
                you must write: "A + B + C <= 0.6".
                While if you specify sublattice_sizes=[2] in the same system per
                primitive cell, to specify the same constraint, write
                "A + B + C <= 1.2" or "0.5 A + 0.5 B + 0.5 C <= 0.6", etc.
            optimize_basis (bool): optional
                Whether to optimize the basis to minimal flip sizes and maximal
                connectivity in the minimum super-cell size.
                When the minimal super-cell size is large, we recommend not to
                optimize basis.
            table_ergodic (bool): optional
                When generating a flip table, whether to add vectors and
                ensure ergodicity under a minimal super-cell size.
                Default to False.
                When the minimal super-cell size is large, we recommend not to
                ensure ergodicity. This is not only because of the computation
                difficulty; but also because at large super-cell size,
                the fraction of inaccessible compositions usually becomes
                minimal.
        """
        self.site_spaces = site_spaces
        self.num_dims = sum(len(species) for species in site_spaces)
        self.dim_ids = get_dim_ids_by_sublattice(self.site_spaces)
        # dimension of "n" format

        # For non-discriminative coordinates
        species = list(set(chain(*self.site_spaces)))
        self.species = []
        for b in species:
            if isinstance(b, Vacancy):
                vac_dupe = False
                for b_old in self.species:
                    if isinstance(b_old, Vacancy):
                        vac_dupe = True
                        break
                if vac_dupe:
                    continue
            self.species.append(b)
        self.species = sorted(self.species)

        species_ids = []
        for species in self.site_spaces:
            sl_dim_ids = []
            for sp in species:
                if not isinstance(sp, Vacancy):
                    sl_dim_ids.append(self.species.index(sp))
                else:
                    for sp2_id, sp2 in enumerate(self.species):
                        if isinstance(sp2, Vacancy):
                            sl_dim_ids.append(sp2_id)
                            break
            species_ids.append(sl_dim_ids)
        self.species_ids = species_ids

        if sublattice_sizes is None:
            self.sublattice_sizes = [1 for _ in range(len(self.site_spaces))]
        elif len(sublattice_sizes) == len(site_spaces):
            self.sublattice_sizes = np.array(sublattice_sizes, dtype=int).tolist()
        else:
            raise ValueError(
                "Sub-lattice number is not the same "
                "in parameters bits and sublattice_sizes."
            )

        self.charge_neutral = charge_neutral
        self.optimize_basis = optimize_basis
        self.table_ergodic = table_ergodic

        # Pre-process input constraints.
        self.other_constraints = other_constraints
        if self.other_constraints is not None:
            self._other_eq_constraints = self.other_constraints["eq"]
            self._other_leq_constraints = self.other_constraints["leq"]
        else:
            self._other_eq_constraints = []
            self._other_leq_constraints = []

        # Set constraint equations An=b (per primitive cell).
        A = []
        b = []
        if charge_neutral:
            A.append([get_oxi_state(sp) for species in site_spaces for sp in species])
            b.append(0)
        for dim_id, sublattice_size in zip(self.dim_ids, self.sublattice_sizes):
            a = np.zeros(self.num_dims, dtype=int)
            a[dim_id] = 1
            A.append(a.tolist())
            b.append(sublattice_size)
        for a, bb in self._other_eq_constraints:
            if len(a) != self.num_dims:
                raise ValueError(
                    f"Constraint length: {len(a)} does not match"
                    f" dimensions: {self.num_dims}!"
                )
            # No-longer enforce integers in a and b.
            # Integerize a.
            a_new, scale = integerize_vector(a)
            # Convert to array before multiplying!!
            A.append(np.round(np.array(a) * scale).astype(int))
            b.append(bb * scale)
        self._A = np.array(A, dtype=int)
        self._b = np.array(b)  # per-prim
        if np.linalg.matrix_rank(self._A) >= self.num_dims:
            raise ValueError("Valid constraints more than number of dimensions!")

        if len(self._other_leq_constraints) > 0:
            self._A_leq = np.array([a for a, bb in self._other_leq_constraints])
            self._b_leq = np.array([bb for a, bb in self._other_leq_constraints])
        else:
            self._A_leq = None
            self._b_leq = None

        self._prim_vertices = None
        # Minimum supercell size required to make prim_vertices coordinates
        # all integer.
        self._min_supercell_size = None
        self._flip_table = None
        self._n0 = None  # n0 base solution at minimum feasible sc size.
        self._vs = None  # Basis vectors, n = n0 + vs.T @ x
        self._comp_grids = {}  # Grid of integer compositions, in "x" format.

    @property
    def prim_vertices(self):
        """Vertex compositions in a primitive cell.

        leq constraints are not considered here.

        Returns:
            prim vertices in "counts" format:
                2D np.ndarray of float
        """
        if self._prim_vertices is None:
            self._prim_vertices = get_nonneg_float_vertices(self._A, self._b)
        return self._prim_vertices

    @property
    def min_supercell_size(self):
        """Minimum super-cell size.

        Computed as the minimum integer that can multiply prim_vertices into
        integral vectors.

        Returns: int
        """
        if self._min_supercell_size is None:
            int_verts, supercell_size = integerize_multiple(self.prim_vertices)
            self._min_supercell_size = supercell_size
        return self._min_supercell_size

    @property
    def num_unconstrained_compositions(self):
        """Estimated number of unconstrained compositions.

        Returns: int
        """
        return np.prod(
            [
                (sublattice_size * self.min_supercell_size) ** len(species)
                for species, sublattice_size in zip(
                    self.site_spaces, self.sublattice_sizes
                )
            ]
        )

    def get_supercell_base_solution(self, supercell_size=None):
        """Find one solution (not natural numbers) in a super-cell size.

        Args:
            supercell_size(int): optional
                Super-cell size in the number of primitive cells.
                If not given, will use self.min_supercell_size.
        Returns:
             1D np.ndarray of int
        """
        # This conserves scalability of grid enumeration, which means a grid
        # enumerated at supercell_size = 2 * a and step = 2 is exactly 2 * the
        # grid enumerated at supercell_size = a and step = 1. But make sure that
        # step = 1 case is called first.
        if supercell_size is None:
            supercell_size = self.min_supercell_size
        # minimum sc size that makes self._b an integer vector.
        # Any allowed supercell_size must be a multiple of it.
        _, min_feasible_supercell_size = integerize_vector(self._b)
        if not supercell_size % min_feasible_supercell_size == 0:
            raise ValueError(
                "Composition constraints can not have any integral "
                f"solution in a super-cell of {supercell_size} prims!"
            )
        if self._n0 is None:
            n0, vs = solve_diophantines(
                self._A, np.round(self._b * min_feasible_supercell_size)
            )
            # n0 and vs must always be ints.
            self._n0 = n0.copy()
        return self._n0 * supercell_size // min_feasible_supercell_size

    @property
    def basis(self):
        """Basis vectors of the integer composition grid.

        if self.optimize_basis is set to true, will be optimized so that flip
        sizes are minimal, and the graph connectivity between enumerated
        compositions in a self.min_supercell_size super-cell is maximized.
        Returns:
            Basis vectors in "counts" format, each in a row:
                2D np.ndarray of int
        """
        if self._vs is None:
            n0, vs = solve_diophantines(
                self._A, np.round(self._b * self.min_supercell_size)
            )
            if self.optimize_basis:
                n_comps = self.num_unconstrained_compositions
                if n_comps > 10**6:
                    warnings.warn(
                        "Basis optimization can be very costly "
                        "at your composition space size = "
                        f"{n_comps}. "
                        "Do this at your own risk!"
                    )
                xs = get_natural_solutions(n0, vs)
                vs_opt = get_optimal_basis(n0, vs, xs)
            else:
                vs_opt = vs.copy()
            self._vs = vs_opt
        return self._vs

    @property
    def flip_table(self):
        """Give flip table vectors.

        If self.table_ergodic is true, will add flips until the flip table is ergodic
        in a self.min_supercell_size super-cell.

        Returns: 2D np.ndarray of int
            Flip vectors in the "counts" format
        """
        if self._flip_table is None:
            if not self.table_ergodic:
                self._flip_table = self.basis.copy()
            else:
                n_comps = self.num_unconstrained_compositions
                if n_comps > 10**6:
                    warnings.warn(
                        "Ergodicity computation can be very costly "
                        "in your composition space, which has "
                        f"{n_comps} unconstrained compositions. "
                        "Do this at your own risk!"
                    )
                n0 = self.get_supercell_base_solution(self.min_supercell_size)
                self._flip_table = get_ergodic_vectors(
                    n0, self.basis, self.min_supercell_grid
                )
        return self._flip_table

    @property
    def flip_reactions(self):
        """Reaction formula representations of table flips.

        Return:
            All reaction formulas (only forward direction):
                list of str
        """
        return [flip_vec_to_reaction(u, self.site_spaces) for u in self.flip_table]

    def get_composition_grid(self, supercell_size=1, step=1):
        """Get the integer compositions ("coordinates").

        Args:
            supercell_size (int):
                Super-cell size to enumerate with.
            step (int): optional
                Step in returning the enumerated compositions.
                If step = N > 1, on each dimension of the composition space,
                we will only yield one composition every N compositions.
                Default to 1.

        Returns: 2D np.ndarray of int
            Integer compositions ("coordinates" format, not normalized):
        """
        # Also scalablity issue.
        scale = None
        supercell_size_prev = None
        step_prev = None
        for k1, k2 in self._comp_grids:
            if (
                supercell_size % k1 == 0
                and step % k2 == 0
                and supercell_size // k1 == step // k2
            ):
                scale = supercell_size // k1
                supercell_size_prev = k1
                step_prev = k2
                break

        if scale is not None:
            return self._comp_grids[(supercell_size_prev, step_prev)] * scale
        else:
            s = np.gcd(supercell_size, step)
            if s > 1:
                return (
                    self.get_composition_grid(
                        supercell_size=supercell_size // s, step=step // s
                    )
                    * s
                )
            else:
                n0 = self.get_supercell_base_solution(supercell_size)
                grid = get_natural_solutions(n0, self.basis, step=step)

                ns = grid @ self.basis + n0  # each row
                if self._A_leq is not None and self._b_leq is not None:
                    _filter_leq = (
                        self._A_leq @ ns.T / supercell_size
                        <= self._b_leq[:, None] + NUM_TOL
                    ).all(axis=0)
                else:
                    _filter_leq = np.ones(len(ns)).astype(bool)

                # Filter inequality constraints.
                self._comp_grids[(supercell_size, step)] = grid[_filter_leq]
                return self._comp_grids[(supercell_size, step)]

    @property
    def min_supercell_grid(self):
        """Get integer compositions on grid at min_supercell_size ("coordinates").

        Returns: 2D np.ndarray of int
            Integer compositions ("coordinates" format, not normalized):
        """
        return self.get_composition_grid(supercell_size=self.min_supercell_size)

    def get_centroid_composition(self, supercell_size=None):
        """Get the closest integer composition to the centroid of polytope.

        Args:
            supercell_size(int): optional
               Super-cell size to get the composition with.
               If not given, will use self.min_supercell_size
        Return: 1D np.ndarray of int
            the closest composition to the centroid ("coordinates" format, not normalized)
        """
        if supercell_size is None:
            supercell_size = self.min_supercell_size
        n0 = self.get_supercell_base_solution(supercell_size)
        return get_natural_centroid(
            n0,
            self.basis,
            supercell_size,
            self._A_leq,
            self._b_leq,
        )

    def translate_format(
        self, c, supercell_size, from_format, to_format="counts", rounding=False
    ):
        """Translate a composition representation to another format.

        Args:
            c (1D ArrayLike or list of Composition):
                Input format. Can be int or fractional.
            supercell_size (int):
                Supercell size of this composition. You must make sure
                it is the right super-cell size of composition c.
            from_format (str):
                Specifies the input format.
                A composition can be expressed in 4 formats:
                1, Count of species on each sub-lattice, ordered and concatenated as
                   each sub-lattice provided in self.bits and self.sublattice_sizes,
                   into an 1D array. ("counts" format)
                2, Coordinates x on the constrained integer grid, computed with
                   n = n0 + V @ x from its counts format representation n,
                   where V are the grid basi vectors and n0 is a base solution in
                   a certain super-cell size. ("coordinates" format).
                3, Composition of each sub-lattice in pymatgen.Composition,
                   but each sub-lattice composition is required to be divided
                   by number of the sites in the sub-lattice, so that the sum of species
                   content can not exceed 1. Vacancies contents are not explicitly
                   included, but taken a 1 - sum(other_species_contents).
                   ("compositions" format).
                4, Count each species in the whole structure, without distinguishing
                   sub-lattice. The species are sorted, and the order of species is
                   given in self.species. ("species-counts" format).
                Note: "species-counts" format cannot be converted to other formats,
                because it does not have information on species distribution across
                sub-lattices.
            to_format (str): optional
                Specified the output format.
                Same as from_format, with an addition "species-counts".
            rounding (bool): optional
                If the returned format is "counts", "coordinates" or "species-counts",
                whether to round up the output array as integers. Default to False.
        Return:
            Depends on the argument "to_format":
                1D np.ndarray of int or float, or list of Composition
        """
        if from_format == "species-counts":
            raise ValueError("species-counts can not be converted to other formats!")

        n = self._convert_to_counts(
            c, form=from_format, supercell_size=supercell_size, rounding=rounding
        )
        return self._convert_counts_to(
            n, form=to_format, supercell_size=supercell_size, rounding=rounding
        )

    def _convert_to_counts(self, c, form, supercell_size, rounding):
        """Convert other composition format to n-format."""
        if form == "counts":
            n = np.array(c)
        elif form == "coordinates":
            n = self.basis.transpose() @ np.array(c) + self.get_supercell_base_solution(
                supercell_size
            )
        elif form == "compositions":
            n = []
            for species, sublattice_size, comp in zip(
                self.site_spaces, self.sublattice_sizes, c
            ):
                if comp.num_atoms > 1 or comp.num_atoms < 0:
                    raise ValueError(
                        f"Sub-lattice composition {c} given in"
                        f" pymatgen.Composition format, but was"
                        f" not normalized to 1!"
                    )
                vac_counted = False
                for specie in species:
                    if isinstance(specie, Vacancy):
                        if vac_counted:
                            raise ValueError(
                                "More than one Vacancy species "
                                "appear in a sub-lattice!"
                            )
                        comp_novac = Composition(
                            {
                                k: v
                                for k, v in comp.items()
                                if not isinstance(k, Vacancy)
                            }
                        )
                        n.append(
                            (1 - comp_novac.num_atoms)
                            * sublattice_size
                            * supercell_size
                        )
                        vac_counted = True
                    else:
                        n.append(comp[specie] * sublattice_size * supercell_size)
            n = np.array(n)
        else:
            raise ValueError(f"Composition format {form} not supported!")

        if rounding:
            n_round = np.array(np.round(n), dtype=int)
            if np.any(np.abs(n_round - n) > NUM_TOL):
                raise ValueError(f"Composition {n} cannot be rounded into integers!")
            n = n_round.copy()

        return n

    def _convert_counts_to(self, n, form, supercell_size, rounding):
        n = np.array(n)
        if np.any(n < -NUM_TOL):
            raise ValueError(f"Composition {n} contains negative species count!")
        if np.any(np.abs(self._A @ (n / supercell_size) - self._b) > NUM_TOL):
            raise ValueError(
                f"Composition {n} violates constraints!"
                f" Numerical tolerance: {NUM_TOL}"
            )

        if form == "counts":
            c = n.copy()
        elif form == "coordinates":
            dn = n - self.get_supercell_base_solution(supercell_size)
            c = np.linalg.pinv(self.basis.T) @ dn
        elif form == "compositions":
            c = []
            for species, sublattice_size, dim_id in zip(
                self.site_spaces, self.sublattice_sizes, self.dim_ids
            ):
                n_sl = n[dim_id] / (sublattice_size * supercell_size)
                c.append(
                    Composition(
                        {
                            sp: n
                            for sp, n in zip(species, n_sl)
                            if not isinstance(sp, Vacancy)
                        }
                    )
                )
        elif form == "species-counts":
            c = np.zeros(len(self.species))
            for dim_id, species_ids in zip(self.dim_ids, self.species_ids):
                c[species_ids] += n[dim_id]
        else:
            raise ValueError(f"Composition format {form} not supported!")

        if rounding and form != "compositions":
            c_round = np.array(np.round(c), dtype=int)
            if np.any(np.abs(c - c_round) > NUM_TOL):
                raise ValueError(f"Composition {c} cannot be rounded into integers!")
            c = c_round.copy()
        return c

    def as_dict(self):
        """Serialize into dictionary.

        Return:
            dict
        """
        bits = [[sp.as_dict() for sp in sl_sps] for sl_sps in self.site_spaces]

        n_cons = len(self.site_spaces)
        if self.charge_neutral:
            n_cons += 1
        eq_constraints = [
            (a, bb, "eq")
            for a, bb in zip(self._A[n_cons:].tolist(), self._b[n_cons:].tolist())
        ]
        leq_constraints = (
            [
                (a, bb, "leq")
                for a, bb in zip(self._A_leq.tolist(), self._b_leq.tolist())
            ]
            if self._A_leq is not None and self._b_leq is not None
            else []
        )

        other_constraints = eq_constraints + leq_constraints

        comp_grids = {
            f"{k[0]}_{k[1]}": v.tolist() for k, v in self._comp_grids.items()
        }  # Encode tuple keys.

        def to_list(arr):
            if arr is not None:
                return np.array(arr).tolist()
            else:
                return None

        return {
            "bits": bits,
            "sublattice_sizes": self.sublattice_sizes,
            "other_constraints": other_constraints,
            "charge_neutral": self.charge_neutral,
            "optimize_basis": self.optimize_basis,
            "table_ergodic": self.table_ergodic,
            "min_supercell_size": self._min_supercell_size,
            "prim_vertices": to_list(self._prim_vertices),
            "n0": to_list(self._n0),
            "vs": to_list(self._vs),
            "flip_table": to_list(self._flip_table),
            "comp_grids": comp_grids,
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
        }

    @classmethod
    def from_dict(cls, d):
        """Load CompositionSpace object from dictionary.

        Args:
            d(dict):
                Dictionary to decode from.
        Return:
            CompositionSpace
        """
        decoder = MontyDecoder()
        bits = [
            [decoder.process_decoded(sp_d) for sp_d in sl_sps] for sl_sps in d["bits"]
        ]
        sublattice_sizes = d.get("sublattice_sizes")
        other_constraints = d.get("other_constraints")
        charge_neutral = d.get("charge_neutral", True)
        optimize_basis = d.get("optimize_basis", False)
        table_ergodic = d.get("table_ergodic", False)

        obj = cls(
            bits,
            sublattice_sizes,
            other_constraints=other_constraints,
            charge_neutral=charge_neutral,
            optimize_basis=optimize_basis,
            table_ergodic=table_ergodic,
        )

        obj._min_supercell_size = d.get("min_supercell_size")

        prim_vertices = d.get("prim_vertices")
        if prim_vertices is not None:
            obj._prim_vertices = np.array(prim_vertices)

        n0 = d.get("n0")
        obj._n0 = np.array(n0, dtype=int) if n0 is not None else None

        vs = d.get("vs")
        if vs is not None:
            obj._vs = np.array(vs, dtype=int)

        flip_table = d.get("flip_table")
        if flip_table is not None:
            obj._flip_table = np.array(flip_table, dtype=int)

        comp_grids = d.get("comp_grids", {})
        comp_grids = {
            (int(k.split("_")[0]), int(k.split("_")[1])): np.array(v, dtype=int)
            for k, v in comp_grids.items()
        }  # Decode tuple keys.
        obj._comp_grids = comp_grids

        return obj



import os

import numpy as np
import numpy.testing as npt
import pytest

from smol.capp.generate.random import _gen_unconstrained_ordered_occu
from smol.moca import SampleContainer, Sampler
from smol.moca.kernel import Metropolis
from smol.moca.kernel.mcusher import Flip, Swap
from tests.utils import assert_pickles

TEMPERATURE = 5000
# Correlations are within ATOL 1E-14, but ewald energies sometimes need more slack
ATOL = 5e-13  # this is still enough precision anyway


@pytest.fixture(params=[1, 5])
def sampler(ensemble, rng, request):
    sampler = Sampler.from_ensemble(
        ensemble,
        temperature=TEMPERATURE,
        seeds=[rng for _ in range(request.param)],
        nwalkers=request.param,
    )
    # fix this additional attribute to sampler to access in gen occus for tests
    sampler.num_sites = ensemble.num_sites
    return sampler


def test_from_ensemble(sampler):
    if hasattr(sampler.samples.metadata, "chemical_potentials"):
        assert isinstance(sampler.mckernels[0]._usher, Flip)
    else:
        assert isinstance(sampler.mckernels[0]._usher, Swap)
    assert isinstance(sampler.mckernels[0], Metropolis)


@pytest.mark.parametrize("thin", (1, 10))
def test_sample(sampler, thin, rng):
    occu = np.vstack(
        [
            _gen_unconstrained_ordered_occu(
                sampler.mckernels[0]._usher.sublattices, rng=rng
            )
            for _ in range(sampler.samples.shape[0])
        ]
    )
    steps = 1000
    samples = [state for state in sampler.sample(1000, occu, thin_by=thin)]
    assert len(samples) == steps // thin

    it = sampler.sample(43, occu, thin_by=7)
    with pytest.warns(RuntimeWarning):
        next(it)


@pytest.mark.parametrize("thin", (1, 10))
def test_run(sampler, thin, rng):
    occu = np.vstack(
        [
            _gen_unconstrained_ordered_occu(kernel._usher.sublattices, rng=rng)
            for kernel in sampler.mckernels
        ]
    )
    steps = 1000
    sampler.run(steps, occu, thin_by=thin)
    assert len(sampler.samples) == steps // thin
    assert 0 <= sampler.efficiency() <= 1

    # pick some random samples and check recorded traces match!
    for i in rng.choice(range(sampler.samples.num_samples), size=50):
        npt.assert_allclose(
            sampler.samples.get_feature_vectors(flat=False)[i],
            np.vstack(
                list(
                    map(
                        sampler.mckernels[0].ensemble.compute_feature_vector,
                        sampler.samples.get_occupancies(flat=False)[i],
                    )
                )
            ),
            atol=ATOL,
        )

    sampler.clear_samples()


def test_anneal(sampler, rng, tmpdir):
    temperatures = np.linspace(2000, 500, 5)
    occu = np.vstack(
        [
            _gen_unconstrained_ordered_occu(
                sampler.mckernels[0]._usher.sublattices, rng=rng
            )
            for _ in range(sampler.samples.shape[0])
        ]
    )
    steps = 100
    sampler.anneal(temperatures, steps, occu)
    expected = []
    for T in temperatures:
        expected += (
            steps
            * sampler.samples.shape[0]
            * [
                T,
            ]
        )
    npt.assert_array_equal(sampler.samples.get_trace_value("temperature"), expected)
    assert sampler.samples.num_samples == len(temperatures) * steps

    # test streaming anneal
    new_container = SampleContainer.from_dict(sampler.samples.as_dict())
    new_container.clear()
    new_sampler = Sampler(sampler.mckernels, new_container)
    file_path = os.path.join(tmpdir, "test.h5")

    new_sampler.anneal(
        temperatures, steps, occu, stream_chunk=10, stream_file=file_path
    )
    assert new_sampler.samples.num_samples == 0
    samples = SampleContainer.from_hdf5(file_path)
    assert samples.num_samples == len(temperatures) * steps
    npt.assert_array_equal(samples.get_trace_value("temperature"), expected)
    os.remove(file_path)
    # test temp error
    with pytest.raises(ValueError):
        sampler.anneal([100, 200], steps)


def test_reshape_occu(ensemble, rng):
    sampler = Sampler.from_ensemble(ensemble, temperature=TEMPERATURE)
    occu = _gen_unconstrained_ordered_occu(ensemble.sublattices, rng=rng)
    assert sampler._reshape_occu(occu).shape == (1, len(occu))


def test_pickles(sampler):
    assert_pickles(sampler)



"""Cython function to compute the change in electrostatic energy from a flips."""

__author__ = "William D. Richards"

cimport numpy as np
import numpy as np


cpdef double delta_ewald_single_flip(const np.int32_t[::1] occu_f,
                                     const np.int32_t[::1] occu_i,
                                     const np.float64_t[:, ::1] ewald_matrix,
                                     const np.int32_t[:, ::1] ewald_indices,
                                     const int site_ind) nogil:
    """Compute the change in electrostatic interaction energy from a flip.

    Args:
        occu_f (ndarray):
            encoded occupancy vector with flip
        occu_i (ndarray):
            encoded occupancy vector without flip
        ewald_matrix (ndarray):
            Ewald matrix for electrostatic interactions
        ewald_indices (ndarray):
            2D array of indices corresponding to a specific site occupation
            in the ewald matrix
        site_ind (int):
            site index for site being flipped

    Returns:
        float: electrostatic interaction energy difference
    """
    cdef int i, j, k, add, sub
    cdef bint ok
    cdef double out = 0
    cdef double out_k

    # values of -1 are vacancies and hence don't have ewald indices
    add = ewald_indices[site_ind, occu_f[site_ind]]
    sub = ewald_indices[site_ind, occu_i[site_ind]]

    # single threaded seems good enough for the majority of systems
    # if multiple threads are ever needed this can be replaced with prange
    for k in range(occu_f.shape[0]):
        i = ewald_indices[k, occu_f[k]]
        out_k = 0
        if i != -1 and add != -1:
            if i != add:
                out_k = out_k + 2 * ewald_matrix[i, add]
            else:
                out_k = out_k + ewald_matrix[i, add]

        j = ewald_indices[k, occu_i[k]]
        if j != -1 and sub != -1:
            if j != sub:
                out_k = out_k - 2 * ewald_matrix[j, sub]
            else:
                out_k = out_k - ewald_matrix[j, sub]
        out += out_k
    return out



"""Get constraints on variables from a processor."""

from numbers import Number
from typing import List, Tuple, Union

import cvxpy as cp
import numpy as np
from cvxpy import Expression
from cvxpy.constraints.constraint import Constraint
from pymatgen.core import Structure

from smol.capp.generate.groundstate.upper_bound.indices import (
    get_variable_indices_for_each_composition_component,
)
from smol.moca.composition.space import CompositionSpace
from smol.moca.ensemble import Sublattice

__author__ = "Fengyu Xie"


def get_normalization_constraints(
    variables: cp.Variable, variable_indices: List[List[int]]
) -> List[Constraint]:
    """Get normalization constraints of variables on each site.

    Args:
        variables(cp.Variable):
            cvxpy variables storing the ground-state result.
        variable_indices(list of lists of int):
            List of variable indices corresponding to each active site index and
            species indices in its site space.
    Returns:
        list of Constraint.
    """
    constraints = []
    for site_indices in variable_indices:
        site_variable_indices = np.array(site_indices, dtype=int)
        site_variable_indices = site_variable_indices[site_variable_indices >= 0]
        if len(site_variable_indices) > 0:
            constraints.append(cp.sum(variables[site_variable_indices]) == 1)
    return constraints


def _extract_constraint_matrix(comp_space, constraint_type):
    """Get the constraint matrix of a specific type."""
    if constraint_type in ["==", "=", "eq"]:
        mat_a = comp_space._A
        vec_b = comp_space._b
        constraint_type = "=="
    elif constraint_type in ["<=", "leq"]:
        mat_a = comp_space._A_leq if comp_space._A_leq is not None else []
        vec_b = comp_space._b_leq if comp_space._b_leq is not None else []
        constraint_type = "<="
    else:
        raise NotImplementedError(f"Constraint type {constraint_type} not supported!")
    return mat_a, vec_b, constraint_type


def _get_formula_constraints(
    comp_space, constraint_type, variables, variables_per_component, skip_cids=None
):
    """Get a specific type of composition constraints."""
    mat_a, vec_b, constraint_type = _extract_constraint_matrix(
        comp_space, constraint_type
    )

    constraints = []
    skip_cids = skip_cids or []
    for cid, (a, b) in enumerate(zip(mat_a, vec_b)):
        # Null constraints.
        if cid in skip_cids:
            continue
        if np.allclose(a, 0):
            if constraint_type == "==" and not np.isclose(b, 0):
                raise ValueError(f"Unsatisfiable constraint an=b, a: {a}, b: {b}.")
            if constraint_type == "<=" and b < 0:
                raise ValueError(f"Unsatisfiable constraint an<=b, a: {a}, b: {b}.")
            continue

        expression = 0
        for dim_id, (indices, n_fixed) in enumerate(variables_per_component):
            # Active sub-lattice.
            if len(indices) > 0 and not np.isclose(a[dim_id], 0):
                expression += cp.sum(variables[indices]) * a[dim_id]
            expression += n_fixed * a[dim_id]
        if not isinstance(expression, Expression):
            if (constraint_type == "==" and expression != b) or (
                constraint_type == "<=" and expression > b
            ):
                raise ValueError(
                    f"Constraint {a} {constraint_type} {b} can never be"
                    f" satisfied because the number of restricted sites"
                    f" can not satisfy this requirement!"
                )
        else:
            if constraint_type == "==":
                constraints.append(expression == b)
            if constraint_type == "<=":
                constraints.append(expression <= b)

    return constraints


def get_composition_space_constraints(
    sublattices: List[Sublattice],
    variables: cp.Variable,
    variable_indices: List[List[int]],
    processor_structure: Structure,
    charge_balanced: bool = True,
    other_constraints: List[
        Union[str, Tuple[Union[dict, List[dict], List[Number]], Number, str]]
    ] = None,
) -> List[Constraint]:
    """Get constraints on species composition in CompositionSpace.

    Supports charge balance and other generic composition constraints.
    See moca.CompositionSpace.
    Args:
        sublattices(list of Sublattice):
            Sub-lattices to build the upper-bound problem on.
        variables(cp.Variable):
            cvxpy variables storing the ground-state result.
        variable_indices(list of lists of int):
            List of variable indices corresponding to each active site index and
            index of species in its site space. Inactive sites will be marked by
            either -1 or -2. See documentation in groundstate.upper_bound.variables.
        processor_structure(Structure):
            The supercell structure stored in a processor's structure attribute.
            The sub-lattices must match the processor structure, or they must be the result
            of splitting with the initial_occupancy. See smol.moca.sublattice for the
            explanation of splitting a sub-lattice.
        charge_balanced(bool): optional
            Whether to enforce charge balance. Default is True.
        other_constraints(list): optional
            Other constraints to set in composition space. See moca.composition
            for detailed description.
            Note that constraints are now to be satisfied with the number of sites
            in sub-lattices of the supercell instead of the primitive cell.
    Returns:
        list of Constraint: Constraints corresponding to the given composition space.
    """
    # Get the variable indices corresponding to each dimension in "counts" format.
    variables_per_component = get_variable_indices_for_each_composition_component(
        sublattices, variable_indices, processor_structure
    )

    # Create the composition space.
    bits = [s.species for s in sublattices]
    sizes = [len(s.sites) for s in sublattices]
    comp_space = CompositionSpace(
        bits,
        sizes,
        charge_neutral=charge_balanced,
        other_constraints=other_constraints,
        optimize_basis=False,
        table_ergodic=False,
    )

    # Charge balance is the first constraint in CompositionSpace.
    n_charge = 1 if comp_space.charge_neutral else 0
    n_sublattices = len(sublattices)
    # Normalization on each sublattice are not necessary because it is already satisfied
    # by normalization on each site.
    skip_cids = list(range(n_charge, n_charge + n_sublattices))
    if np.allclose(comp_space._A[0, :], 0) and np.isclose(comp_space._b[0], 0):
        # Neutral alloy, no need to constrain charge.
        skip_cids = [0] + skip_cids

    # EQ, LEQ and GEQ. No skip in LEQ and GEQ.
    constraints = []
    constraints.extend(
        _get_formula_constraints(
            comp_space, "eq", variables, variables_per_component, skip_cids
        )
    )
    constraints.extend(
        _get_formula_constraints(comp_space, "leq", variables, variables_per_component)
    )
    # All inequality constraints must be of leq type.

    return constraints


def get_fixed_composition_constraints(
    sublattices: List[Sublattice],
    variables: cp.Variable,
    variable_indices: List[List[int]],
    processor_structure: Structure,
    fixed_composition: List[int],
) -> List[Constraint]:
    """Fix the count of species in the super-cell.

    Used for searching ground-states in a canonical ensemble.
    Args:
        sublattices(list of Sublattice):
            Sub-lattices to build the upper-bound problem on.
        variables(cp.Variable):
            cvxpy variables storing the ground-state result.
        variable_indices(list of lists of int):
            List of variable indices corresponding to each active site index and
            index of species in its site space. Inactive sites will be marked by
            either -1 or -2. See documentation in groundstate.upper_bound.variables.
        processor_structure(Structure):
            The supercell structure stored in a processor's structure attribute.
            The sub-lattices must match the processor structure, or they must be the result
            of splitting with the initial_occupancy. See smol.moca.sublattice for the
            explanation of splitting a sub-lattice.
        fixed_composition(list of int):
            Amount of each species to be fixed in the SUPER-cell, in CompositionSpace
            "counts" format. You are fully responsible for setting the order of species
            in the list correctly as would be generated by a CompositionSpace.
    Return:
        list of Constraint: Constraints corresponding to a fixed composition on each
        sub-lattice.
    """
    # Get the variable indices corresponding to each dimension in "counts" format.
    variables_per_component = get_variable_indices_for_each_composition_component(
        sublattices, variable_indices, processor_structure
    )

    constraints = []
    for dim_id, (indices, n_fixed) in enumerate(variables_per_component):
        expression = 0
        if len(indices) > 0:
            expression += cp.sum(variables[indices])
        expression += n_fixed

        if n_fixed > fixed_composition[dim_id]:
            raise ValueError(
                f"Fixed composition {fixed_composition} can not"
                f" be satisfied because the {dim_id}'th component"
                f" is always occupied by {n_fixed} species, more than"
                f" the number allowed!"
            )
        if not isinstance(expression, Expression):
            if expression != fixed_composition[dim_id]:
                raise ValueError(
                    f"Fixed composition {fixed_composition} can not"
                    f" be satisfied because the {dim_id}'th component"
                    f" is restricted to be occupied by {expression} species,"
                    f" not equals to the number required!"
                )
        else:
            constraints.append(expression == fixed_composition[dim_id])

    return constraints



"""Utilities to perform groundstate tests. Only valid for solver_test ensemble."""

import numpy as np
import numpy.testing as npt
from pymatgen.core import Species

from smol.capp.generate.groundstate.upper_bound.indices import (
    get_sublattice_indices_by_site,
)
from smol.capp.generate.groundstate.upper_bound.variables import (
    get_variable_values_from_occupancy,
)
from smol.cofe.space.domain import Vacancy
from smol.moca.kernel.mcusher import Swap


def get_random_solver_test_occu(sublattices):
    # Only use this on orig_sublattices.
    # Li+ 3 Mn2+ 3 Mn4+ 2 Ti4+ 1 Vac 3 O2- 9 F- 3
    cation_sublattice = None
    anion_sublattice = None
    for sublattice in sublattices:
        if Species("Li", 1) in sublattice.species:
            cation_sublattice = sublattice
        if Species("O", -2) in sublattice.species:
            anion_sublattice = sublattice

    li_va_sites = np.random.choice(cation_sublattice.sites, size=6, replace=False)
    li_sites = np.random.choice(li_va_sites, size=3, replace=False)
    va_sites = np.setdiff1d(li_va_sites, li_sites)
    mn_ti_sites = np.setdiff1d(cation_sublattice.sites, li_va_sites)
    ti_sites = np.random.choice(mn_ti_sites, size=1, replace=False)
    mn_sites = np.setdiff1d(mn_ti_sites, ti_sites)
    mn2_sites = np.random.choice(mn_sites, size=3, replace=False)
    mn4_sites = np.setdiff1d(mn_sites, mn2_sites)

    o_sites = np.random.choice(anion_sublattice.sites, size=9, replace=False)
    o2_sites = o_sites
    f_sites = np.setdiff1d(anion_sublattice.sites, o_sites)

    li_code = cation_sublattice.encoding[
        cation_sublattice.species.index(Species("Li", 1))
    ]
    mn2_code = cation_sublattice.encoding[
        cation_sublattice.species.index(Species("Mn", 2))
    ]
    mn4_code = cation_sublattice.encoding[
        cation_sublattice.species.index(Species("Mn", 4))
    ]
    ti_code = cation_sublattice.encoding[
        cation_sublattice.species.index(Species("Ti", 4))
    ]
    va_code = cation_sublattice.encoding[cation_sublattice.species.index(Vacancy())]
    o2_code = anion_sublattice.encoding[
        anion_sublattice.species.index(Species("O", -2))
    ]
    f_code = anion_sublattice.encoding[anion_sublattice.species.index(Species("F", -1))]

    occu = np.zeros(24, dtype=np.int32) - 1
    occu[li_sites] = li_code
    occu[mn2_sites] = mn2_code
    occu[mn4_sites] = mn4_code
    occu[ti_sites] = ti_code
    occu[va_sites] = va_code
    occu[o2_sites] = o2_code
    occu[f_sites] = f_code

    assert np.all(occu >= 0)

    return occu.astype(np.int32)


def get_random_variable_values(sublattices):
    # Not always charge balanced.
    num_sites = sum(len(s.sites) for s in sublattices)
    site_sublattice_ids = get_sublattice_indices_by_site(sublattices)

    values = []
    for site_id in range(num_sites):
        sl_id = site_sublattice_ids[site_id]
        sublattice = sublattices[sl_id]
        if len(sublattice.species) > 1 and site_id in sublattice.active_sites:
            site_vals = [0 for _ in range(len(sublattice.species))]
            site_vals[np.random.choice(len(sublattice.species))] = 1
            values.extend(site_vals)

    return np.array(values, dtype=int)


def get_random_neutral_occupancy(
    sublattices, initial_occupancy, canonical=False, force_flip=False
):
    # Guarantee charge balanced.
    ti_code = None
    mn4_code = None
    ti_sites = None
    mn4_sites = None
    for sublattice in sublattices:
        if Species("Ti", 4) in sublattice.species:
            ti_code = sublattice.encoding[sublattice.species.index(Species("Ti", 4))]
            mn4_code = sublattice.encoding[sublattice.species.index(Species("Mn", 4))]
            ti_sites = sublattice.sites[initial_occupancy[sublattice.sites] == ti_code]
            mn4_sites = sublattice.sites[
                initial_occupancy[sublattice.sites] == mn4_code
            ]
    if canonical:
        threshold = 1.1
    elif not force_flip:
        threshold = 0.5
    else:
        threshold = -0.1

    if np.random.random() > threshold:
        # Ti-Mn flip that would not change charge.
        if np.random.random() > 0.5:
            flip = [(np.random.choice(ti_sites), mn4_code)]
        else:
            flip = [(np.random.choice(mn4_sites), ti_code)]
    else:
        swapper = Swap(sublattices)
        flip = swapper.propose_step(initial_occupancy)

    # Apply that flip.
    rand_occu = initial_occupancy.copy()
    for site_id, code in flip:
        rand_occu[site_id] = code
    return rand_occu.astype(np.int32)


def get_random_neutral_variable_values(
    sublattices, initial_occupancy, variable_indices, canonical=False, force_flip=False
):
    # Guarantee charge balanced.
    rand_occu = get_random_neutral_occupancy(
        sublattices, initial_occupancy, canonical=canonical, force_flip=force_flip
    )
    return get_variable_values_from_occupancy(sublattices, rand_occu, variable_indices)


def validate_correlations_from_occupancy(expansion_processor, occupancy):
    # Check whether our interpretation of corr function is correct.
    # Enforce int32.
    occupancy = np.array(occupancy, dtype=np.int32)
    space = expansion_processor.cluster_subspace
    sc_matrix = expansion_processor.supercell_matrix
    mappings = space.supercell_orbit_mappings(sc_matrix)

    corr = np.zeros(space.num_corr_functions)
    corr[0] = 1
    for orbit, mapping in zip(space.orbits, mappings):
        # Use un-flatten version now for easier access.
        n = orbit.bit_id
        corr_tensors = orbit.correlation_tensors
        n_bit_combos = corr_tensors.shape[0]  # index of bit combos
        mapping = np.array(mapping, dtype=int)
        n_clusters = mapping.shape[0]  # cluster image index
        for bid in range(n_bit_combos):
            for cid in range(n_clusters):
                cluster_state = occupancy[mapping[cid]].tolist()
                corr[n] += corr_tensors[tuple((bid, *cluster_state))]
            corr[n] /= n_clusters
            n += 1
    npt.assert_array_almost_equal(
        corr * expansion_processor.size,
        expansion_processor.compute_feature_vector(occupancy),
    )


def validate_interactions_from_occupancy(decomposition_processor, occupancy):
    # Check whether our interpretation of corr function is correct.
    orbit_tensors = decomposition_processor._interaction_tensors
    # Enforce int32.
    occupancy = np.array(occupancy, dtype=np.int32)
    space = decomposition_processor.cluster_subspace
    sc_matrix = decomposition_processor.supercell_matrix
    mappings = space.supercell_orbit_mappings(sc_matrix)

    corr = np.zeros(decomposition_processor.cluster_subspace.num_orbits)
    corr[0] = orbit_tensors[0]
    n = 1
    for mapping in mappings:
        # Use un-flatten version now for easier access.
        mapping = np.array(mapping, dtype=int)
        n_clusters = mapping.shape[0]  # cluster image index
        for cid in range(n_clusters):
            cluster_state = occupancy[mapping[cid]].tolist()
            corr[n] += orbit_tensors[n][tuple(cluster_state)]
        corr[n] /= n_clusters
        n += 1
    npt.assert_array_almost_equal(
        corr * decomposition_processor.size,
        decomposition_processor.compute_feature_vector(occupancy),
    )


def evaluate_correlations_from_variable_values(grouped_terms, variable_values):
    # Evaluate correlation functions or interactions from variable values.
    variable_values = np.array(variable_values, dtype=int)
    corr = []
    for group in grouped_terms:
        f = 0
        for var_inds, corr_factor, _ in group:
            if len(var_inds) == 0:
                f += corr_factor
            else:
                f += corr_factor * np.prod(variable_values[var_inds])
        corr.append(f)
    return np.array(corr)



from itertools import product
from random import choice, choices

import numpy as np
import numpy.testing as npt
import pytest

from smol.capp.generate.random import _gen_unconstrained_ordered_occu
from smol.constants import kB
from smol.moca.kernel import (
    Metropolis,
    MulticellMetropolis,
    UniformlyRandom,
    WangLandau,
)
from smol.moca.kernel.base import ALL_MCUSHERS, ThermalKernelMixin
from smol.moca.kernel.bias import FugacityBias
from smol.moca.kernel.mcusher import Flip, Swap, TableFlip
from smol.moca.trace import StepTrace, Trace
from tests.utils import assert_pickles

kernels_with_bias = [UniformlyRandom, Metropolis]
kernels_no_bias = [MulticellMetropolis, WangLandau]
kernels = kernels_with_bias + kernels_no_bias
ushers = ALL_MCUSHERS


@pytest.fixture(params=product(kernels, ushers), scope="module")
def mckernel(ensemble, rng, request):
    kwargs = {}
    kernel_class, step_type = request.param

    if issubclass(kernel_class, ThermalKernelMixin):
        kwargs["temperature"] = 5000

    if kernel_class == WangLandau:
        kwargs["min_enthalpy"] = -5000
        kwargs["max_enthalpy"] = 5000
        kwargs["bin_size"] = 10
        kwargs["check_period"] = 5

    if step_type == "MultiStep":
        kwargs["mcusher"] = choice(
            [c for c in ushers if c not in ("MultiStep", "Composite")]
        )
        kwargs["step_lengths"] = 4
    elif step_type == "Composite":
        kwargs["mcushers"] = choices(
            [c for c in ushers if c not in ("MultiStep", "Composite")], k=2
        )
    # TODO create ensemble with different cell shapes
    if kernel_class == MulticellMetropolis:
        kernels = [
            Metropolis(ensemble, step_type=step_type, **kwargs) for _ in range(10)
        ]
        kernel = kernel_class(
            kernels, kwargs["temperature"]
        )  # , kernel_hop_periods=101)
        kernel.set_aux_state(
            np.vstack(
                [
                    _gen_unconstrained_ordered_occu(kernel.mcusher.sublattices, rng=rng)
                    for _ in range(len(kernel.mckernels))
                ]
            )
        )
    else:
        kernel = kernel_class(ensemble, step_type=step_type, **kwargs)
        kernel.set_aux_state(
            _gen_unconstrained_ordered_occu(kernel.mcusher.sublattices, rng=rng)
        )
    return kernel


@pytest.fixture(params=product(kernels_with_bias, ushers), scope="module")
def mckernel_bias(ensemble, request):
    kwargs = {}
    kernel_class, step_type = request.param
    if issubclass(kernel_class, ThermalKernelMixin):
        kwargs["temperature"] = 5000

    if step_type == "MultiStep":
        kwargs["mcusher"] = choice(
            [c for c in ushers if c not in ("MultiStep", "Composite")]
        )
        kwargs["step_lengths"] = 4
    elif step_type == "Composite":
        kwargs["mcushers"] = choices(
            [c for c in ushers if c not in ("MultiStep", "Composite")], k=2
        )

    kernel = kernel_class(ensemble, step_type=step_type, **kwargs)
    kernel.bias = FugacityBias(kernel.mcusher.sublattices)
    return kernel


@pytest.mark.parametrize(
    "step_type, mcusher", [("swap", Swap), ("flip", Flip), ("table-flip", TableFlip)]
)
def test_constructor(ensemble, step_type, mcusher):
    kernel = Metropolis(ensemble, step_type=step_type, temperature=500)
    assert isinstance(kernel.mcusher, mcusher)
    assert isinstance(kernel.trace, StepTrace)
    assert "temperature" in kernel.trace.names
    kernel.bias = FugacityBias(kernel.mcusher.sublattices)
    assert "bias" in kernel.trace.delta_trace.names


def test_single_step(mckernel, rng):
    occu = _gen_unconstrained_ordered_occu(mckernel.mcusher.sublattices, rng=rng)
    mckernel.set_aux_state(occu)
    prev_kernel_index = 0  # for multicell metropolis
    prev_occu = occu.copy()

    for _ in range(100):
        trace = mckernel.single_step(occu)
        curr_features = mckernel.ensemble.compute_feature_vector(occu)

        if trace.accepted:
            npt.assert_array_equal(trace.occupancy, occu)
            if (
                isinstance(mckernel, MulticellMetropolis)
                and prev_kernel_index != trace.kernel_index
            ):
                prev_features = mckernel._features[prev_kernel_index]
                assert np.allclose(
                    mckernel.mckernels[
                        prev_kernel_index
                    ].ensemble.compute_feature_vector(prev_occu),
                    prev_features,
                )
                prev_kernel_index = trace.kernel_index
            else:
                prev_features = mckernel.ensemble.compute_feature_vector(prev_occu)

            npt.assert_allclose(
                curr_features - prev_features,
                trace.delta_trace.features,
                rtol=1e-5,
                atol=1e-8,
            )
        else:
            npt.assert_array_equal(trace.occupancy, prev_occu)

        if isinstance(mckernel, MulticellMetropolis):
            npt.assert_allclose(
                mckernel._features[trace.kernel_index],
                curr_features,
                rtol=1e-5,
                atol=1e-8,
            )
            assert mckernel.trace.kernel_index == mckernel._current_kernel_index
            assert (
                mckernel.trace
                is mckernel.mckernels[mckernel._current_kernel_index].trace
            )
        elif isinstance(mckernel, WangLandau):
            assert mckernel.bin_size > 0
            assert isinstance(mckernel.levels, np.ndarray)
            assert isinstance(mckernel.entropy, np.ndarray)
            assert isinstance(mckernel.histogram, np.ndarray)

            assert "histogram" in trace.names
            assert "occurrences" in trace.names
            assert "entropy" in trace.names
            assert "cumulative_mean_features" in trace.names
            assert "mod_factor" in trace.names

        # save previous occu
        prev_occu[:] = occu


def test_single_step_bias(mckernel_bias, rng):
    mckernel_bias.set_aux_state(
        _gen_unconstrained_ordered_occu(mckernel_bias.mcusher.sublattices)
    )
    for _ in range(100):
        occu = _gen_unconstrained_ordered_occu(
            mckernel_bias.mcusher.sublattices, rng=rng
        )
        trace = mckernel_bias.single_step(occu.copy())
        # assert delta bias is there and recorded
        assert isinstance(trace.delta_trace.bias, np.ndarray)
        assert len(trace.delta_trace.bias.shape) == 0  # 0 dimensional
        if trace.accepted:
            assert not np.array_equal(trace.occupancy, occu)
        else:
            npt.assert_array_equal(trace.occupancy, occu)


def test_temperature_setter(single_canonical_ensemble):
    metropolis_kernel = Metropolis(
        single_canonical_ensemble, step_type="flip", temperature=500
    )
    assert metropolis_kernel.beta == 1 / (kB * metropolis_kernel.temperature)
    metropolis_kernel.temperature = 500
    assert metropolis_kernel.beta == 1 / (kB * 500)


def test_trace(rng):
    trace = Trace(first=np.ones(10), second=np.zeros(10))
    assert all(isinstance(val, np.ndarray) for _, val in trace.items())

    trace.third = rng.random(10)
    assert all(isinstance(val, np.ndarray) for _, val in trace.items())
    names = ["first", "second", "third"]
    assert all(name in names for name in trace.names)

    with pytest.raises(TypeError):
        trace.fourth = "blabla"
        Trace(one=np.zeros(40), two=66)

    steptrace = StepTrace(one=np.zeros(10))
    assert isinstance(steptrace.delta_trace, Trace)
    with pytest.raises(ValueError):
        steptrace.delta_trace = np.ones(10)
        StepTrace(delta_trace=np.ones(10))

    with pytest.raises(TypeError):
        steptrace.fourth = "blabla"
        StepTrace(one=np.zeros(40), two=66)

    # check saving
    assert trace.as_dict() == trace.__dict__
    steptrace_d = steptrace.__dict__.copy()
    steptrace_d["delta_trace"] = steptrace_d["delta_trace"].__dict__.copy()
    assert steptrace.as_dict() == steptrace_d


def test_pickles(mckernel):
    assert_pickles(mckernel)


# TODO add direct multicell tests, especially to check proper updates of traces
#  before/after cell jumps



"""Test utilities for managing composition constraints."""

import numpy.testing as npt
import pytest
from pymatgen.core import DummySpecies, Element, Species

# To test manager.
from smol.cofe.space.domain import Vacancy
from smol.moca.composition.constraints import (
    convert_constraint_string,
    handle_side_string,
)
from smol.moca.composition.space import CompositionSpace


def test_handle_side_string():
    # Good expressions.
    ag1 = Species("Ag", 1)
    mn = Element("Mn")
    mn3 = Species("Mn", 3)
    cu = Element("Cu")
    vac = Vacancy()
    x = DummySpecies("X")

    cases = [
        ("Ag+ + 2 Mn", ([(1, ag1, None), (2, mn, None)], 0)),
        (" Ag+ +  2 Mn ", ([(1, ag1, None), (2, mn, None)], 0)),
        ("- 1 Ag+ +3.5 Mn3+", ([(-1, ag1, None), (3.5, mn3, None)], 0)),
        ("- Ag+ + 3.5 Mn3+", ([(-1, ag1, None), (3.5, mn3, None)], 0)),
        ("Ag+(0) +2 Mn3+", ([(1, ag1, 0), (2, mn3, None)], 0)),
        ("Ag+(00) +2 Mn3+(10) + 1.5", ([(1, ag1, 0), (2, mn3, 10)], 1.5)),
        (
            "Vacancy(0) + Mn3+(10) - 4.1 Cu",
            ([(1, vac, 0), (1, mn3, 10), (-4.1, cu, None)], 0),
        ),
        ("X(100) + 2", ([(1, x, 100)], 2)),
    ]

    for s, standard in cases:
        assert handle_side_string(s) == standard

    # Bad test cases.
    # species string not separated from number in the front.
    with pytest.raises(Exception):
        _ = handle_side_string("Ag+ + 2Mn")
    # species string not separated from operator in the front.
    with pytest.raises(Exception):
        _ = handle_side_string("Ag+ +Mn")
    # Space exist between species and its sub-lattice index label.
    with pytest.raises(Exception):
        _ = handle_side_string("Ag+ (0) +Mn")
    # Space inserted within a species string
    with pytest.raises(Exception):
        _ = handle_side_string("Ag+, spin=1(0) +Mn")


@pytest.fixture
def bits():
    return [
        [Species("Li", 1), Species("Mn", 3), Species("Ti", 4), Vacancy()],
        [Species("Li", 1), Species("O", -2), Species("O", -1), Species("F", -1)],
    ]


def test_convert_constraint_string(bits):
    cases = [
        ("- Li+ +3 O2-  <= - 10.2", ([-1, 0, 0, 0, -1, 3, 0, 0], -10.2, "<=")),
        ("- Li+(0) + O2- = 2", ([-1, 0, 0, 0, 0, 1, 0, 0], 2, "=")),
        (
            "- Li+(0) + 5 O- - 1.5 >= 4 Li+(1) - Mn3+ + Vacancy(0)",
            ([-1, 1, 0, -1, -4, 0, 5, 0], 1.5, ">="),
        ),
        ("- Li+ + 3 F- == 2 Ti4+", ([-1, 0, -2, 0, -1, 0, 0, 3], 0, "==")),
    ]

    for s, standard in cases:
        assert standard == convert_constraint_string(s, bits)

    # Bad cases where symbols are not properly separated by spaces.
    with pytest.raises(ValueError):
        _ = convert_constraint_string("- Li+ +3 O2- <=- 10.2", bits)
    with pytest.raises(ValueError):
        _ = convert_constraint_string("- Li+ +3 O2->= - 10.2", bits)
    # Bad case when symbol is not allowed:
    with pytest.raises(ValueError):
        _ = convert_constraint_string("- Li+ +3 O2- > - 10.2", bits)
    with pytest.raises(ValueError):
        _ = convert_constraint_string("- Li+ +3 O2->= - 10.2", bits)
    # Bad case when species can not be found on any sub-lattice
    with pytest.raises(ValueError):
        _ = convert_constraint_string("- K+(0) +3 O2- <= - 10.2", bits)
    with pytest.raises(ValueError):
        _ = convert_constraint_string("- K+ +3 O2- <= - 10.2", bits)


def test_constraints_manager(bits):
    sublattice_sizes = [10, 6]

    # Charge goes first, then normalizations and others.
    a_eq_std = [
        [1, 3, 4, 0, 1, -2, -1, -1],
        [1, 1, 1, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 1, 1, 1],
        [-2, 0, 1, 0, -2, 0, 0, 0],
        [1, 1, 1, 2, 0, 4, -1, 3],
    ]
    b_eq_std = [0, 10, 6, 6, 4]
    a_leq_std = [[1, 0.5, 0, 2, 0, 0, -1, 3], [0, 0, 1, -1, 0.4, 0, 0, 0]]
    b_leq_std = [10, 0]
    # No more geq constraints stored in manager.

    # Test all allowed formats of input.
    other_constraints = [
        # converted to int.
        "-1 Li+ + 0.5 Ti4+ == 3",
        ([1, 1, 1, 2, 0, 4, -1, 3], 4, "="),
        # not converted to int.
        "Li+(0) + 0.5 Mn3+(0) + 2 Vacancy(0) - O-(1) + 3 F-(1) <= 10",
        "- Ti4+ + Vacancy >= 0.4 Li+(1)",
    ]

    comp_space = CompositionSpace(
        bits,
        sublattice_sizes,
        charge_neutral=True,
        other_constraints=other_constraints,
    )

    npt.assert_array_equal(a_eq_std, comp_space._A)
    npt.assert_array_equal(b_eq_std, comp_space._b)
    npt.assert_array_equal(a_leq_std, comp_space._A_leq)
    npt.assert_array_equal(b_leq_std, comp_space._b_leq)

    # Test None cases.
    other_constraints = [
        # converted to int.
        "-1 Li+ + 0.5 Ti4+ == 3",
        ([1, 1, 1, 2, 0, 4, -1, 3], 4, "="),
    ]
    comp_space = CompositionSpace(
        bits,
        sublattice_sizes,
        charge_neutral=True,
        other_constraints=other_constraints,
    )
    npt.assert_array_equal(a_eq_std, comp_space._A)
    npt.assert_array_equal(b_eq_std, comp_space._b)
    assert comp_space._A_leq is None
    assert comp_space._b_leq is None

    comp_space = CompositionSpace(
        bits,
        sublattice_sizes,
        charge_neutral=True,
    )
    npt.assert_array_equal(a_eq_std[:-2], comp_space._A)
    npt.assert_array_equal(b_eq_std[:-2], comp_space._b)
    assert comp_space._A_leq is None
    assert comp_space._b_leq is None



"""Implementation of Thermodynamic Ensemble for MC sampling."""

__author__ = "Luis Barroso-Luque"

from collections import Counter

import numpy as np
from monty.json import MSONable, jsanitize
from pymatgen.core.composition import ChemicalPotential

from smol.cofe.space.domain import get_species
from smol.moca.processor import (
    ClusterDecompositionProcessor,
    ClusterExpansionProcessor,
    CompositeProcessor,
    EwaldProcessor,
)
from smol.moca.processor.base import Processor
from smol.moca.sublattice import Sublattice


class ChemicalPotentialManager:
    """Chemical potential descriptor for use Ensemble class."""

    natural_parameter: float = -1.0

    def __set_name__(self, owner, name):
        """Set the private variable names."""
        self.public_name = name
        self.private_name = "_" + name

    def __get__(self, obj, objtype=None):
        """Return the chemical potentials if set None otherwise."""
        value = getattr(obj, self.private_name, None)
        return value if value is None else value["value"]

    def __set__(self, obj, value):
        """Set the table given the owner and value."""
        if value is None:  # call delete if set to None
            self.__delete__(obj)
            return

        for spec, count in Counter(map(get_species, value.keys())).items():
            if count > 1:
                raise ValueError(
                    f"{count} values of the chemical potential for the same "
                    f"species {spec} were provided.\n Make sure the dictionary "
                    "you are using has only string keys or only Species "
                    "objects as keys."
                )
        value = {
            get_species(k): v for k, v in value.items() if get_species(k) in obj.species
        }
        if set(value.keys()) != set(obj.species):
            raise ValueError(
                "Chemical potentials given are missing species. "
                "Values must be given for each of the following:"
                f" {obj.species}"
            )

        # if first instantiation concatenate the natural parameter
        if not hasattr(obj, self.private_name):
            obj.natural_parameters = np.append(
                obj.natural_parameters, self.natural_parameter
            )
        setattr(
            obj,
            self.private_name,
            {"value": ChemicalPotential(value), "table": self._build_table(obj, value)},
        )
        # update the ensemble dictionary and _boundaries list
        if hasattr(obj, "thermo_boundaries"):
            obj.thermo_boundaries.update({self.public_name: value})
        else:
            setattr(obj, "thermo_boundaries", {self.public_name: value})

    def __delete__(self, obj):
        """Delete the boundary condition."""
        if hasattr(obj, self.private_name):
            del obj.__dict__[self.private_name]
        if (
            hasattr(obj, "thermo_boundaries")
            and self.public_name in obj.thermo_boundaries
        ):
            del obj.thermo_boundaries[self.public_name]
        if obj.num_energy_coefs < len(obj.natural_parameters):
            obj.natural_parameters = obj.natural_parameters[:-1]  # remove last entry

    @staticmethod
    def _build_table(obj, value):
        """Set the chemical potentials and update table."""
        num_cols = max(max(sl.encoding) for sl in obj.sublattices) + 1

        # Sublattice can only be initialized as default, or split from default.
        table = np.zeros((obj.num_sites, num_cols))
        for sublatt in obj.active_sublattices:
            ordered_pots = [value[sp] for sp in sublatt.site_space]
            table[sublatt.sites[:, None], sublatt.encoding] = ordered_pots
        return table


class Ensemble(MSONable):
    """Thermodynamic ensemble class.

    Attributes:
        thermo_boundaries (dict):
            Dictionary with corresponding thermodynamic boundaries, i.e.
            chemical potentials. This is kept only for descriptive purposes.
    """

    chemical_potentials = ChemicalPotentialManager()

    def __init__(self, processor, sublattices=None, chemical_potentials=None):
        """Initialize class instance.

        Args:
            processor (Processor):
                a processor that can compute the change in property given
                a set of flips.
            sublattices (list of Sublattice): optional
                list of Sublattice objects representing sites in the processor
                supercell with same site spaces.
        """
        if sublattices is None:
            sublattices = processor.get_sublattices()
        self.thermo_boundaries = {}  # not pretty way to save general info
        self._params = processor.coefs  # natural parameters
        self._processor = processor
        self._sublattices = sublattices
        self.chemical_potentials = chemical_potentials

    @classmethod
    def from_cluster_expansion(
        cls,
        cluster_expansion,
        supercell_matrix,
        processor_type="decomposition",
        **kwargs,
    ):
        """Initialize an ensemble from a cluster expansion.

        Convenience constructor to instantiate an ensemble. This will take
        care of initializing the correct processor based on the
        ClusterExpansion.

        Args:
            cluster_expansion (ClusterExpansion):
                A cluster expansion object.
            supercell_matrix (ndarray):
                Supercell matrix defining the system size.
            processor_type(str): optional
                Type of processor to be used besides external term.
                Can use "decomposition" (default) or "expansion".
            **kwargs:
                Keyword arguments to pass to ensemble constructor. Such as
                sublattices, sublattice_probabilities, chemical_potentials,
                fugacity_fractions.

        Returns:
            Ensemble
        """
        if len(cluster_expansion.cluster_subspace.external_terms) > 0:
            processor = CompositeProcessor(
                cluster_expansion.cluster_subspace, supercell_matrix
            )
            if processor_type == "decomposition":
                ceprocessor = ClusterDecompositionProcessor(
                    cluster_expansion.cluster_subspace,
                    supercell_matrix,
                    cluster_expansion.cluster_interaction_tensors,
                )
            elif processor_type == "expansion":
                ceprocessor = ClusterExpansionProcessor(
                    cluster_expansion.cluster_subspace,
                    supercell_matrix,
                    cluster_expansion.coefs[:-1],
                )
            else:
                raise ValueError(f"Processor type {processor_type}" f" not supported!")
            processor.add_processor(ceprocessor)
            # at some point determine term and spinup processor maybe with a
            # factory, if we ever implement more external terms.
            ewald_term = cluster_expansion.cluster_subspace.external_terms[0]
            ewprocessor = EwaldProcessor(
                cluster_expansion.cluster_subspace,
                supercell_matrix,
                ewald_term=ewald_term,
                coefficient=cluster_expansion.coefs[-1],
            )
            processor.add_processor(ewprocessor)
        else:
            if processor_type == "decomposition":
                processor = ClusterDecompositionProcessor(
                    cluster_expansion.cluster_subspace,
                    supercell_matrix,
                    cluster_expansion.cluster_interaction_tensors,
                )
            elif processor_type == "expansion":
                processor = ClusterExpansionProcessor(
                    cluster_expansion.cluster_subspace,
                    supercell_matrix,
                    cluster_expansion.coefs,
                )
            else:
                raise ValueError(f"Processor type {processor_type}" f" not supported!")
        return cls(processor, **kwargs)

    @property
    def num_sites(self):
        """Get the total number of sites in the supercell."""
        return self.processor.num_sites

    @property
    def num_energy_coefs(self):
        """Return the number of coefficients used in the energy expansion."""
        return len(self._processor.coefs)

    @property
    def system_size(self):
        """Get size of supercell in number of prims."""
        return self.processor.size

    @property
    def processor(self):
        """Get the ensemble processor."""
        return self._processor

    # TODO make a setter for these that checks sublattices are correct and
    #  all sites are included.
    @property
    def sublattices(self):
        """Get list of Sublattices included in ensemble."""
        return self._sublattices

    @property
    def active_sublattices(self):
        """Get list of active sub-lattices."""
        return [s for s in self.sublattices if s.is_active]

    @property
    def restricted_sites(self):
        """Get indices of all restricted sites."""
        return np.concatenate(
            [sublattice.restricted_sites for sublattice in self.sublattices]
        )

    @property
    def species(self):
        """Species on active sublattices.

        These are minimal species required in setting chemical potentials.
        """
        return list(
            {sp for sublatt in self.active_sublattices for sp in sublatt.site_space}
        )

    @property
    def natural_parameters(self):
        """Get the vector of natural parameters.

        The natural parameters correspond to the fit coefficients of the
        underlying processor plus any additional terms involved in the Legendre
        transformation corresponding to the ensemble.
        """
        return self._params

    @natural_parameters.setter
    def natural_parameters(self, value):
        """Set the value of the natural parameters.

        Should only allow appending to the original energy coefficients
        """
        if not np.array_equal(self.processor.coefs, value[: self.num_energy_coefs]):
            raise ValueError("The original expansion coefficients can not be changed!")
        self._params = value

    def split_sublattice_by_species(self, sublattice_id, occu, species_in_partitions):
        """Split a sub-lattice in system by its occupied species.

        An example use case might be simulating topotactic Li extraction
        and insertion, where we want to consider Li/Vac, TM and O as
        different sub-lattices that can not be mixed by swapping.

        Args:
            sublattice_id (int):
                The index of sub-lattice to split in self.sublattices.
            occu (np.ndarray[int]):
                An occupancy array to reference with.
            species_in_partitions (List[List[int|Species|Vacancy|Element|str]]):
                Each sub-list contains a few species or encodings of species in
                the site space to be grouped as a new sub-lattice, namely,
                sites with occu[sites] == specie in the sub-list, will be
                used to initialize a new sub-lattice.
                Sub-lists will be pre-sorted to ascending order.
        """
        splits = self.sublattices[sublattice_id].split_by_species(
            occu, species_in_partitions
        )
        self._sublattices = (
            self._sublattices[:sublattice_id]
            + splits
            + self._sublattices[sublattice_id + 1 :]
        )
        if self.chemical_potentials is not None:
            # Species in active sub-lattices may change after split.
            # Need to reset and rebuild chemical potentials.
            chemical_potentials = {
                spec: self.chemical_potentials[spec] for spec in self.species
            }
            self.chemical_potentials = chemical_potentials

    def compute_feature_vector(self, occupancy):
        """Compute the feature vector for a given occupancy.

        The feature vector includes the necessary features required to compute
        the exponent determining the relative probability for the given
        occupancy (i.e. a generalized enthalpy). The feature vector for
        ensembles represents the sufficient statistics.

        For a cluster expansion, the feature vector is the
        correlation vector x system size

        Args:
            occupancy (ndarray[np.int32]):
                encoded occupancy string

        Returns:
            ndarray: feature vector
        """
        # Don't enforce int32 here as it may increase overhead.
        features = self.processor.compute_feature_vector(occupancy)

        if self.chemical_potentials is not None:
            chemical_work = sum(
                self._chemical_potentials["table"][site][species]
                for site, species in enumerate(occupancy)
            )
            features = np.append(features, chemical_work)

        return features

    def compute_feature_vector_change(self, occupancy, step):
        """Compute the change in the feature vector from a given step.

        Args:
            occupancy (ndarray[np.int32]):
                encoded occupancy string.
            step (list of tuple):
                a sequence of flips given by MCUsher.propose_step

        Returns:
            ndarray: difference in feature vector
        """
        # Don't enforce int32 here as it may increase overhead.
        delta_features = self.processor.compute_feature_vector_change(occupancy, step)

        if self.chemical_potentials is not None:
            delta_work = sum(
                self._chemical_potentials["table"][f[0]][f[1]]
                - self._chemical_potentials["table"][f[0]][occupancy[f[0]]]
                for f in step
            )
            delta_features = np.append(delta_features, delta_work)

        return delta_features

    def restrict_sites(self, sites):
        """Restricts (freezes) the given sites.

        This will exclude those sites from being flipped during a Monte Carlo
        run. If some of the given indices refer to inactive sites, there will
        be no effect.

        Args:
            sites (Sequence):
                indices of sites in the occupancy string to restrict.
        """
        for sublattice in self.sublattices:
            sublattice.restrict_sites(sites)

    def reset_restricted_sites(self):
        """Unfreeze all previously restricted sites."""
        for sublattice in self.sublattices:
            sublattice.reset_restricted_sites()

    def as_dict(self):
        """Get Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        ensemble_d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "thermo_boundaries": jsanitize(self.thermo_boundaries),
            "processor": self._processor.as_dict(),
            "sublattices": [s.as_dict() for s in self._sublattices],
        }
        return ensemble_d

    @classmethod
    def from_dict(cls, ensemble_d):
        """Instantiate a CanonicalEnsemble from dict representation.

        Args:
            ensemble_d (dict):
                dictionary representation.
        Returns:
            CanonicalEnsemble
        """
        ensemble = cls(
            Processor.from_dict(ensemble_d["processor"]),
            [Sublattice.from_dict(s) for s in ensemble_d["sublattices"]],
        )
        chemical_potentials = ensemble_d["thermo_boundaries"].get("chemical_potentials")
        if chemical_potentials is not None:
            ensemble.chemical_potentials = chemical_potentials

        return ensemble
